from shiny import *

app_ui = ui.page_fluid(
    ui.input_checkbox("somevalue", "Some value", False), ui.output_ui("value")
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    def value():
        return input.somevalue()


app = App(app_ui, server)



from shiny import *

app_ui = ui.page_fluid(
    ui.input_text("caption", "Caption:", "Data summary"),
    ui.output_text_verbatim("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.text
    def value():
        return input.caption()


app = App(app_ui, server)



__all__ = (
    "modal_button",
    "modal",
    "modal_show",
    "modal_remove",
)

import sys
from typing import Optional, Union

from shiny.types import MISSING, MISSING_TYPE

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from htmltools import tags, Tag, div, HTML, TagChildArg, TagAttrArg

from .._docstring import add_example
from ..session import Session, require_active_session


def modal_button(
    label: TagChildArg, icon: TagChildArg = None, **kwargs: TagChildArg
) -> Tag:
    """
    Creates a button that will dismiss a :func:`modal` (useful when customising the
    ``footer`` of :func:`modal`).

    Parameters
    ----------
    label
        An input label.
    icon
        An icon to appear inline with the button/link.
    kwargs
        Attributes to be applied to the button.

    Returns
    -------
    A UI element

    See Also
    -------
    ~shiny.ui.modal
    ~shiny.ui.modal_show
    ~shiny.ui.modal_remove

    Example
    -------
    See :func:`modal`.
    """
    return tags.button(
        icon,
        label,
        {"class": "btn btn-default"},
        type="button",
        data_dismiss="modal",
        data_bs_dismiss="modal",
        **kwargs
    )


@add_example()
def modal(
    *args: TagChildArg,
    title: Optional[str] = None,
    footer: Union[TagChildArg, MISSING_TYPE] = MISSING,
    size: Literal["m", "s", "l", "xl"] = "m",
    easy_close: bool = False,
    fade: bool = True,
    **kwargs: TagAttrArg
) -> Tag:
    """
    Creates the UI for a modal dialog, using Bootstrap's modal class. Modals are
    typically used for showing important messages, or for presenting UI that requires
    input from the user, such as a user name and/or password input.

    Parameters
    ----------
    args
        UI elements for the body of the modal.
    title
        An optional title for the modal dialog.
    footer
        UI for footer. Use ``None`` for no footer.
    size
        One of "s" for small, "m" (the default) for medium, or "l" for large.
    easy_close
        If ``True``, the modal dialog can be dismissed by clicking outside the dialog
        box, or be pressing the Escape key. If ``False`` (the default), the modal dialog
        can't be dismissed in those ways; instead it must be dismissed by clicking on a
        ``modal_button()``, or from a call to ``modal_remove()`` on the server.
    fade
        If ``False``, the modal dialog will have no fade-in animation (it will simply
        appear rather than fade in to view).
    kwargs
        Attributes to be applied to the modal's body tag.

    Returns
    -------
    A UI element

    See Also
    -------
    ~shiny.ui.modal_show
    ~shiny.ui.modal_remove
    ~shiny.ui.modal_button
    """

    title_div = None
    if title:
        title_div = div(tags.h4(title, class_="modal-title"), class_="modal-header")

    if isinstance(footer, MISSING_TYPE):
        footer = modal_button("Dismiss")
    if footer is not None:
        footer = div(footer, class_="modal-footer")

    dialog = div(
        div(
            title_div,
            div({"class": "modal-body"}, *args, **kwargs),
            footer,
            class_="modal-content",
        ),
        class_="modal-dialog"
        + ({"s": " modal-sm", "l": " modal-lg", "xl": " modal-xl"}.get(size, "")),
    )

    # jQuery plugin doesn't work in Bootstrap 5, but vanilla JS doesn't work in Bootstrap 4 :sob:
    js = "\n".join(
        [
            "if (window.bootstrap && !window.bootstrap.Modal.VERSION.match(/^4\\. /)) {",
            "  var modal=new bootstrap.Modal(document.getElementById('shiny-modal'))",
            "  modal.show()",
            "} else {",
            "  $('#shiny-modal').modal().focus()",
            "}",
        ]
    )

    backdrop = None if easy_close else "static"
    keyboard = None if easy_close else "false"

    return div(
        dialog,
        tags.script(HTML(js)),
        id="shiny-modal",
        class_="modal fade" if fade else "modal",
        tabindex="-1",
        data_backdrop=backdrop,
        data_bs_backdrop=backdrop,
        data_keyboard=keyboard,
        data_bs_keyboard=keyboard,
    )


def modal_show(modal: Tag, session: Optional[Session] = None) -> None:
    """
    Show a modal dialog.

    Parameters
    ----------
    modal
        Typically a :func:`modal` instance.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    See Also
    -------
    ~shiny.ui.modal_remove
    ~shiny.ui.modal

    Example
    -------
    See :func:`modal`.
    """
    session = require_active_session(session)
    msg = session._process_ui(modal)
    session._send_message_sync({"modal": {"type": "show", "message": msg}})


def modal_remove(session: Optional[Session] = None) -> None:
    """
    Remove a modal dialog.

    Parameters
    ----------
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    See Also
    -------
    ~shiny.ui.modal_show
    ~shiny.ui.modal

    Example
    -------
    See :func:`modal`.
    """
    session = require_active_session(session)
    session._send_message_sync({"modal": {"type": "remove", "message": None}})



__all__ = ("download_button", "download_link")

from typing import Optional

from htmltools import tags, Tag, TagChildArg, TagAttrArg, css

from .._docstring import add_example
from .._namespaces import resolve_id
from .._shinyenv import is_pyodide


@add_example()
def download_button(
    id: str,
    label: TagChildArg,
    icon: TagChildArg = None,
    width: Optional[str] = None,
    **kwargs: TagAttrArg,
) -> Tag:
    """
    Create a download button

    Parameters
    ----------
    id
        An id for the download.
    label
        An input label.
    icon
        An icon to display on the button.
    width
        The width of the button.
    kwargs
        Additional attributes for the button.

    Returns
    -------
    A UI element

    See Also
    --------
    ~shiny.Session.download
    ~shiny.ui.download_link
    """

    return tags.a(
        icon,
        label,
        {"class": "btn btn-default shiny-download-link", "style": css(width=width)},
        id=resolve_id(id),
        # This is a fake link that just results in a 404. It will be replaced by a
        # working link after the server side logic runs, so this link will only be
        # visited in cases where the user clicks the button too fast, or if the server
        # never defines a handler for this download button.
        href="session/0/download/missing_download",
        target="_blank",
        # We can't use `download` in pyodide mode, because the browser chooses not to
        # route the download through the service worker in that case. (Observed by
        # jcheng on 1/7/2022, using Chrome Version 96.0.4664.110.)
        download=None if is_pyodide else True,
        **kwargs,
    )


@add_example()
def download_link(
    id: str,
    label: TagChildArg,
    icon: TagChildArg = None,
    width: Optional[str] = None,
    **kwargs: TagAttrArg,
) -> Tag:
    """
    Create a download button

    Parameters
    ----------
    id
        An id for the download.
    label
        An input label.
    icon
        An icon to display on the button.
    width
        The width of the button.
    kwargs
        Additional attributes for the button.

    Returns
    -------
    A UI element

    See Also
    --------
    ~shiny.Session.download
    ~shiny.ui.download_link
    """

    return tags.a(
        icon,
        label,
        {"class": "shiny-download-link", "style": css(width=width)},
        id=resolve_id(id),
        href="session/0/download/missing_download",
        target="_blank",
        download=None if is_pyodide else True,
        **kwargs,
    )



# Needed for NotRequired with Python 3.7 - 3.9
# See https://www.python.org/dev/peps/pep-0655/#usage-in-python-3-11
from __future__ import annotations

__all__ = (
    "MISSING",
    "MISSING_TYPE",
    "FileInfo",
    "ImgData",
    "SafeException",
    "SilentException",
    "SilentCancelOutputException",
)

import sys
from typing import Union, Optional, Tuple, Dict, Any

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import NotRequired, TypedDict
else:
    from typing_extensions import NotRequired, TypedDict

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from htmltools import TagChildArg

from ._docstring import add_example

# Sentinel value - indicates a missing value in a function call.
class MISSING_TYPE:
    pass


MISSING: MISSING_TYPE = MISSING_TYPE()


# Information about a single file, with a structure like:
#   {'name': 'mtcars.csv', 'size': 1303, 'type': 'text/csv', 'datapath: '/...../mtcars.csv'}
# The incoming data doesn't include 'datapath'; that field is added by the
# FileUploadOperation class.
class FileInfo(TypedDict):
    """
    Information about a file upload.

    See Also
    --------
    ~shiny.ui.input_file

    Example
    -------
    See :func:`~shiny.ui.input_file`.
    """

    name: str
    """The name of the file."""
    size: int
    """The size of the file in bytes."""
    type: str
    """The MIME type of the file."""
    datapath: str
    """The path to the file on the server."""


class ImgData(TypedDict):
    """
    Return type for :func:`~shiny.render.image`.

    See Also
    --------
    ~shiny.render.image

    Example
    -------
    See :func:`~shiny.render.image`.
    """

    src: str
    """The ``src`` attribute of the ``<img>`` tag."""
    width: NotRequired[Union[str, float]]
    """The ``width`` attribute of the ``<img>`` tag."""
    height: NotRequired[Union[str, float]]
    """The ``height`` attribute of the ``<img>`` tag."""
    alt: NotRequired[str]
    """The ``alt`` attribute of the ``<img>`` tag."""
    style: NotRequired[str]
    """The ``style`` attribute of the ``<img>`` tag."""


@add_example()
class SafeException(Exception):
    """
    Throw a safe exception.

    When ``shiny.App.SANITIZE_ERRORS`` is ``True`` (which is the case
    in some production environments like RStudio Connect), exceptions are sanitized
    to prevent leaking of sensitive information. This class provides a way to
    generate an error that is OK to be displayed to the user.
    """

    pass


@add_example()
class SilentException(Exception):
    """
    Throw a silent exception.

    Normally, when an exception occurs inside a reactive context, it's either:

    - Displayed to the user (as a big red error message)
        - This happens when the exception is raised from an output context (e.g., :func:`shiny.render.ui`)
    - Crashes the application
        - This happens when the exception is raised from an :func:`shiny.reactive.Effect`

    This exception is used to silently throw inside a reactive context, meaning that
    execution is paused, and no output is shown to users (or the python console).

    See Also
    --------
    ~SilentCancelOutputException
    """

    pass


@add_example()
class SilentCancelOutputException(Exception):
    """
    Throw a silent exception and don't clear output

    Similar to :class:`~SilentException`, but if thrown in an output context,
    existing output isn't cleared.

    See Also
    --------
    ~SilentException
    """

    pass


class ActionButtonValue(int):
    pass


class NavSetArg(Protocol):
    """
    An value suitable for passing to a navigation container (e.g.,
    :func:`~shiny.ui.navset_tab`).
    """

    def resolve(
        self, selected: Optional[str], context: Dict[str, Any]
    ) -> Tuple[TagChildArg, TagChildArg]:
        """
        Resolve information provided by the navigation container.

        Parameters
        ----------
        selected
            The value of the navigation item to be shown on page load.
        context
            Additional context supplied by the navigation container.
        """
        ...

    def get_value(self) -> Optional[str]:
        """
        Get the value of this navigation item (if any).

        This value is only used to determine what navigation item should be shown
        by default when none is specified (i.e., the first navigation item that
        returns a value is used to determine the container's ``selected`` value).
        """
        ...



from shiny import *
from datetime import date

app_ui = ui.page_fluid(
    ui.input_date("date1", "Date:", value="2016-02-29"),
    # Default value is the date in client's time zone
    ui.input_date("date2", "Date:"),
    # value is always yyyy-mm-dd, even if the display format is different
    ui.input_date("date3", "Date:", value="2016-02-29", format="mm/dd/yy"),
    # Pass in a Date object
    ui.input_date("date4", "Date:", value=date(2016, 2, 29)),
    # Use different language and different first day of week
    ui.input_date("date5", "Date:", language="ru", weekstart=1),
    # Start with decade view instead of default month view
    ui.input_date("date6", "Date:", startview="decade"),
    # Disable Mondays and Tuesdays.
    ui.input_date("date7", "Date:", daysofweekdisabled=[1, 2]),
    # Disable specific dates.
    ui.input_date(
        "date8",
        "Date:",
        value="2016-02-29",
        datesdisabled=["2016-03-01", "2016-03-02"],
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



from shiny import *

app_ui = ui.page_fluid(
    ui.layout_sidebar(
        ui.panel_sidebar(
            ui.input_slider("controller", "Controller", min=1, max=3, value=1)
        ),
        ui.panel_main(
            ui.navset_tab_card(
                ui.nav("Panel 1", "Panel 1 content", value="panel1"),
                ui.nav("Panel 2", "Panel 2 content", value="panel2"),
                ui.nav("Panel 3", "Panel 3 content", value="panel3"),
                id="inTabset",
            )
        ),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        ui.update_navs("inTabset", selected="panel" + str(input.controller()))


app = App(app_ui, server)



import asyncio
import contextlib
import typing
import unittest.mock


class MockTime:
    """Patches time.monotonic() and asyncio.sleep(), replacing actual time with
    simulated time. The intention is to allow time-based code to be unit tested faster
    than realtime.

    Note that this class doesn't currently patch other asyncio time-based functions like
    wait_for(), wait(), as_completed() (although I think we certainly could--we just
    don't need them at this time).
    """

    def __init__(self, now: float = 0):
        self._time = now
        # The tuple elements are 1) time to wake, 2) tiebreaking int, 3) event to be
        # signaled upon wake. The tiebreaking int is necessary to make sorting possible
        # when two sleepers have the same time to wake, as asyncio.Event instances
        # cannot be compared to each other.
        self._sleepers: typing.List[typing.Tuple[float, int, asyncio.Event]] = []
        self._is_advancing = False
        self._i = 0

    @contextlib.contextmanager
    def __call__(self):
        with unittest.mock.patch("time.monotonic", new=self._monotonic):
            with unittest.mock.patch("asyncio.sleep", new=self._sleep):
                yield

    def _monotonic(self):
        return self._time

    async def advance_time(self, secs: float):
        if self._is_advancing:
            raise RuntimeError("MockTime.advance was called reentrantly")

        self._is_advancing = True
        try:
            if secs < 0:
                raise ValueError("MockTime cannot move backwards in time")
            end_time = self._time + secs

            await yield_event_loop()

            # It's OK for self._sleepers to be mutated as we go
            while len(self._sleepers) > 0 and self._sleepers[0][0] <= end_time:
                self._time = self._sleepers[0][0]
                self._sleepers.pop(0)[2].set()

                # Give just-awakened task a chance to run.
                await yield_event_loop()
            self._time = end_time

            await yield_event_loop()

        finally:
            self._is_advancing = False

    async def _sleep(self, delay: float) -> None:
        delay = max(0, delay)
        target = self._time + delay
        wake = asyncio.Event()
        self._sleepers.append((target, self._i, wake))
        self._i += 1
        # Oldest first
        self._sleepers.sort()
        await wake.wait()


async def yield_event_loop():
    # Ordinarily we'd just call asyncio.sleep(0), but we've mocked it, so we have to do
    # something lower-level.

    f: asyncio.Future[None] = asyncio.Future()

    def done():
        f.set_result(None)

    asyncio.get_event_loop().call_soon(done)
    await f



"""
Tools for reactively rendering output for the user interface.
"""

from ._render import (  # noqa: F401
    RenderFunction,  # pyright: ignore[reportUnusedImport]
    RenderFunctionAsync,  # pyright: ignore[reportUnusedImport]
    RenderText,  # pyright: ignore[reportUnusedImport]
    RenderTextAsync,  # pyright: ignore[reportUnusedImport]
    text,
    RenderPlot,  # pyright: ignore[reportUnusedImport]
    RenderPlotAsync,  # pyright: ignore[reportUnusedImport]
    plot,
    RenderImage,  # pyright: ignore[reportUnusedImport]
    RenderImageAsync,  # pyright: ignore[reportUnusedImport]
    image,
    RenderTable,  # pyright: ignore[reportUnusedImport]
    RenderTableAsync,  # pyright: ignore[reportUnusedImport]
    table,
    RenderUI,  # pyright: ignore[reportUnusedImport]
    RenderUIAsync,  # pyright: ignore[reportUnusedImport]
    ui,
)


__all__ = (
    "text",
    "plot",
    "image",
    "table",
    "ui",
)



from shiny import *

app_ui = ui.page_fluid(
    ui.input_slider("controller", "Controller", min=0, max=20, value=10),
    ui.input_numeric("inNumber", "Input number", 0),
    ui.input_numeric("inNumber2", "Input number 2", 0),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        x = input.controller()
        ui.update_numeric("inNumber", value=x)
        ui.update_numeric(
            "inNumber2",
            label="Number label " + str(x),
            value=x,
            min=x - 10,
            max=x + 10,
            step=5,
        )


app = App(app_ui, server)



__all__ = ("insert_ui", "remove_ui")

import sys
from typing import Optional

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from htmltools import TagChildArg

from .._docstring import add_example
from ..session import Session, require_active_session


@add_example()
def insert_ui(
    ui: TagChildArg,
    selector: str,
    where: Literal["beforeBegin", "afterBegin", "beforeEnd", "afterEnd"] = "beforeEnd",
    multiple: bool = False,
    immediate: bool = False,
    session: Optional[Session] = None,
) -> None:
    """
    Insert UI objects

    Parameters
    ----------
    ui
        The UI object you want to insert. This can be anything that you usually put
        inside your apps's ui function. If you're inserting multiple elements in one
        call, make sure to wrap them in either a :func:`~shiny.ui.TagList` or a
        :func:`~shiny.ui.tags.div` (the latter option has the advantage that you can
        give it an id to make it easier to reference or remove it later on). If you want
        to insert raw html, use :func:`~shiny.ui.HTML`.
    selector
        A string that is accepted by jQuery's selector (i.e. the string ``s`` to be
        placed in a ``$(s)`` jQuery call) which determines the element(s) relative to
        which you want to insert your UI object.
    where
        Where your UI object should go relative to the selector:

        - beforeBegin: Before the selector element itself
        - afterBegin: Just inside the selector element, before its first child
        - beforeEnd: Just inside the selector element, after its last child (default)
        - afterEnd: After the selector element itself

        Adapted from
        https://developer.mozilla.org/en-US/docs/Web/API/Element/insertAdjacentHTML.
    multiple
        In case your selector matches more than one element, multiple determines whether
        Shiny should insert the UI object relative to all matched elements or just
        relative to the first matched element (default).
    immediate
        Whether the UI object should be immediately inserted or removed, or whether
        Shiny should wait until all outputs have been updated and all effects have been
        run (default).
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    Note
    ----
    This function allows you to dynamically add arbitrary UI into your app, whenever you
    want, as many times as you want. Unlike :func:`~shiny.render.ui`, the UI generated
    with `insert_ui` is persistent: once it's created, it stays there until removed by
    :func:`remove_ui`. Each new call to `insert_ui` creates more UI objects, in addition
    to the ones already there (all independent from one another). To update a part of
    the UI (ex: an input object), you must use the appropriate render function or a
    customized reactive function.

    See Also
    --------
    ~shiny.ui.remove_ui
    ~shiny.render.ui
    """

    session = require_active_session(session)

    def callback() -> None:
        session._send_insert_ui(
            selector=selector,
            multiple=multiple,
            where=where,
            content=session._process_ui(ui),
        )

    if immediate:
        callback()
    else:
        session.on_flushed(callback, once=True)


@add_example()
def remove_ui(
    selector: str,
    multiple: bool = False,
    immediate: bool = False,
    session: Optional[Session] = None,
) -> None:
    """
    Remove UI objects

    Parameters
    ----------
    selector
        A string that is accepted by jQuery's selector (i.e. the string ``x`` to be
        placed in a ``$(x)`` jQuery call) which determines the element(s) to remove. If
        you want to remove a Shiny input or output, note that many of these are wrapped
        in ``<div>``s, so you may need to use a somewhat complex selector — see the
        Examples below. (Alternatively, you could also wrap the inputs/outputs that you
        want to be able to remove easily in a ``<div>`` with an id.)
    multiple
        In case your selector matches more than one element, multiple determines whether
        Shiny should insert the UI object relative to all matched elements or just
        relative to the first matched element (default).
    immediate
        Whether the UI object should be immediately inserted or removed, or whether
        Shiny should wait until all outputs have been updated and all effects have been
        run (default).
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    See Also
    -------
    ~shiny.ui.insert_ui
    ~shiny.render.ui
    """

    session = require_active_session(session)

    def callback():
        session._send_remove_ui(selector=selector, multiple=multiple)

    if immediate:
        callback()
    else:
        session.on_flushed(callback, once=True)



"""A package for building reactive web applications."""

__version__ = "0.2.1"

from ._shinyenv import is_pyodide as _is_pyodide

# User-facing subpackages that should be available on `from shiny import *`
from . import reactive
from . import render
from .session import (
    Session,
    Inputs,
    Outputs,
)
from . import session
from . import ui

# Private submodules that have some user-facing functionality
from ._app import App
from ._validation import req
from ._deprecated import *

from . import module

if _is_pyodide:
    # In pyodide, avoid importing _main because it imports packages that aren't
    # available.
    run_app = None
else:
    from ._main import run_app


# N.B.: we intentionally don't import 'developer-facing' submodules (e.g.,
# html_dependencies) so that they aren't super visible when you `import shiny`, but
# developers who know what they're doing can import them directly.


__all__ = (
    # public sub-packages
    "reactive",
    "render",
    "session",
    "ui",
    # _app.py
    "App",
    # _main.py
    "run_app",
    # _modules.py
    "module",
    # _session.py
    "Session",
    "Inputs",
    "Outputs",
    # _validation.py
    "req",
    # _deprecated.py
    "render_text",
    "render_plot",
    "render_image",
    "render_ui",
    "event",
)



__all__ = (
    "output_plot",
    "output_image",
    "output_text",
    "output_text_verbatim",
    "output_table",
    "output_ui",
)

from typing import Optional
from htmltools import tags, Tag, div, css, TagAttrArg, TagFunction

from .._docstring import add_example
from .._namespaces import resolve_id


@add_example()
def output_plot(
    id: str, width: str = "100%", height: str = "400px", inline: bool = False
) -> Tag:
    """
    Create a output container for a static plot.

    Place a :func:`~shiny.render.plot` result in the user interface. See
    :func:`~shiny.render.plot` for more details on what types of plots are supported.

    Parameters
    ----------
    id
        An input id.
    width
        The CSS width, e.g. '400px', or '100%'
    height
        The CSS height, e.g. '100%' or '600px'
    inline
        If ``True``, the result is displayed inline

    Returns
    -------
    A UI element

    See Also
    -------
    ~shiny.render.plot
    ~shiny.ui.output_image
    """
    res = output_image(id=resolve_id(id), width=width, height=height, inline=inline)
    res.add_class("shiny-plot-output")
    return res


@add_example()
def output_image(
    id: str, width: str = "100%", height: str = "400px", inline: bool = False
) -> Tag:
    """
    Create a output container for a static image.

    Parameters
    ----------
    id
        An input id.
    width
        The CSS width, e.g. '400px', or '100%'
    height
        The CSS height, e.g. '100%' or '600px'
    inline
        If ``True``, the result is displayed inline

    Returns
    -------
    A UI element

    See Also
    -------
    ~shiny.render.image
    ~shiny.ui.output_plot
    """
    func = tags.span if inline else div
    style = None if inline else css(width=width, height=height)
    return func(id=resolve_id(id), class_="shiny-image-output", style=style)


@add_example()
def output_text(
    id: str, inline: bool = False, container: Optional[TagFunction] = None
) -> Tag:
    """
    Create a output container for some text.

    Parameters
    ----------
    id
        An input id.
    inline
        If ``True``, the result is displayed inline
    container
        A Callable that returns the output container.

    Returns
    -------
    A UI element

    Note
    ----
    Text is HTML-escaped prior to rendering.

    See Also
    -------
    ~shiny.render.text
    ~shiny.ui.output_text_verbatim
    """

    if not container:
        container = tags.span if inline else tags.div
    return container(id=resolve_id(id), class_="shiny-text-output")


def output_text_verbatim(id: str, placeholder: bool = False) -> Tag:
    """
    Create a output container for some text.

    Place a :func:`~shiny.render.text` result in the user interface.
    Differs from :func:`~shiny.ui.output_text` in that it wraps the text in a
    fixed-width container with a gray-ish background color and border.

    Parameters
    ----------
    id
        An input id.
    placeholder
        If the output is empty or ``None``, should an empty rectangle be displayed to
        serve as a placeholder? (does not affect behavior when the output is nonempty)

    Returns
    -------
    A UI element

    See Also
    -------
    ~shiny.render.text
    ~shiny.ui.output_text

    Example
    -------
    See :func:`~shiny.ui.output_text`
    """

    cls = "shiny-text-output" + (" noplaceholder" if not placeholder else "")
    return tags.pre(id=resolve_id(id), class_=cls)


@add_example()
def output_table(id: str, **kwargs: TagAttrArg) -> Tag:
    return tags.div({"class": "shiny-html-output"}, id=resolve_id(id), **kwargs)


@add_example()
def output_ui(
    id: str,
    inline: bool = False,
    container: Optional[TagFunction] = None,
    **kwargs: TagAttrArg
) -> Tag:
    """
    Create a output container for a UI (i.e., HTML) element.

    Parameters
    ----------
    id
        An input id.
    inline
        If ``True``, the result is displayed inline
    container
        A Callable that returns the output container.
    kwargs
        Attributes to be applied to the output container.

    Returns
    -------
    A UI element

    See Also
    -------
    ~shiny.render.ui
    ~shiny.ui.output_text
    """

    if not container:
        container = tags.span if inline else tags.div
    return container({"class": "shiny-html-output"}, id=resolve_id(id), **kwargs)



from typing import Optional, Union

from htmltools import (
    tags,
    Tag,
    TagList,
    TagChildArg,
    TagChild,
    HTMLDependency,
    head_content,
)

from ..types import MISSING, MISSING_TYPE


def shiny_input_label(id: str, label: TagChildArg = None) -> Tag:
    cls = "control-label" + ("" if label else " shiny-label-null")
    return tags.label(label, class_=cls, id=id + "-label", for_=id)


def get_window_title(
    title: Optional[Union[str, Tag, TagList]],
    window_title: Union[str, MISSING_TYPE] = MISSING,
) -> Optional[HTMLDependency]:
    if title is not None and isinstance(window_title, MISSING_TYPE):
        window_title = _find_child_strings(title)

    if isinstance(window_title, MISSING_TYPE):
        return None
    else:
        return head_content(tags.title(window_title))


def _find_child_strings(x: Union[Tag, TagList, TagChild]) -> str:
    if isinstance(x, Tag) and x.name not in ("script", "style"):
        x = x.children
    if isinstance(x, TagList):
        strings = [_find_child_strings(y) for y in x]
        return " ".join(filter(lambda x: x != "", strings))
    if isinstance(x, str):
        return x
    return ""



from shiny import *

app_ui = ui.page_fluid(
    ui.input_slider("controller", "Controller", min=0, max=20, value=10),
    ui.input_text("inText", "Input text"),
    ui.input_text("inText2", "Input text 2"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        x = str(input.controller())
        # This will change the value of input$inText, based on x
        ui.update_text("inText", value="New text " + x)
        # Can also set the label, this time for input$inText2
        ui.update_text("inText2", label="New label " + x, value="New text" + x)


app = App(app_ui, server)



from shiny import *

app_ui = ui.page_fluid(
    ui.tags.p("The first radio button group controls the second"),
    ui.input_radio_buttons(
        "inRadioButtons", "Input radio buttons", ["Item A", "Item B", "Item C"]
    ),
    ui.input_radio_buttons(
        "inRadioButtons2", "Input radio buttons 2", ["Item A", "Item B", "Item C"]
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        x = input.inRadioButtons()

        # Can also set the label and select items
        ui.update_radio_buttons(
            "inRadioButtons2",
            label="Radio buttons label " + x,
            choices=[x],
            selected=x,
        )


app = App(app_ui, server)



import copy
import contextlib
import os
import secrets
from typing import Any, List, Union, Dict, Callable, cast, Optional

from htmltools import Tag, TagList, HTMLDocument, HTMLDependency, RenderedHTML

import starlette.applications
import starlette.exceptions
import starlette.middleware
import starlette.routing
import starlette.websockets
from starlette.requests import Request
from starlette.responses import Response, HTMLResponse, JSONResponse
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from ._autoreload import autoreload_url, InjectAutoreloadMiddleware
from ._connection import Connection, StarletteConnection
from ._shinyenv import is_pyodide
from ._error import ErrorMiddleware
from .html_dependencies import require_deps, jquery_deps, shiny_deps
from .http_staticfiles import StaticFiles
from .reactive import on_flushed
from .session import Inputs, Outputs, Session, session_context


# Default values for App options.
LIB_PREFIX: str = "lib/"
SANITIZE_ERRORS: bool = False
SANITIZE_ERROR_MSG: str = "An error has occurred. Check your logs or contact the app author for clarification."


class App:
    """
    Create a Shiny app instance.

    Parameters
    ----------
    ui
        The UI definition for the app (e.g., a call to :func:`~shiny.ui.page_fluid`
        with nested controls).
    server
        A function which is called once for each session, ensuring that each app is
        independent.
    static_assets
        An absolute directory containing static files to be served by the app.
    debug
        Whether to enable debug mode.

    Example
    -------

    .. code-block:: python

        from shiny import *

        app_ui = ui.page_fluid("Hello Shiny!")

        def server(input: Inputs, output: Outputs, session: Session):
            pass

        app = App(app_ui, server)
    """

    lib_prefix: str = "lib/"
    """
    A path prefix to place before all HTML dependencies processed by
    ``register_web_dependency()``.
    """

    sanitize_errors: bool = False
    """
    Whether or not to show a generic message (``SANITIZE_ERRORS=True``) or the actual
    message (``SANITIZE_ERRORS=False``) in the app UI when an error occurs. This flag
    may default to ``True`` in some production environments (e.g., RStudio Connect).
    """

    sanitize_error_msg: str = "An error has occurred. Check your logs or contact the app author for clarification."
    """
    The message to show when an error occurs and ``SANITIZE_ERRORS=True``.
    """

    ui: RenderedHTML
    server: Callable[[Inputs, Outputs, Session], None]

    def __init__(
        self,
        ui: Union[Tag, TagList],
        server: Optional[Callable[[Inputs, Outputs, Session], None]],
        *,
        static_assets: Optional[Union[str, "os.PathLike[str]"]] = None,
        debug: bool = False,
    ) -> None:
        self.ui: RenderedHTML = _render_page(ui, lib_prefix=self.lib_prefix)

        if server is None:

            def _server(inputs: Inputs, outputs: Outputs, session: Session):
                pass

            server = _server

        self.server = server

        self._debug: bool = debug

        # Settings that the user can change after creating the App object.
        self.lib_prefix: str = LIB_PREFIX
        self.sanitize_errors: bool = SANITIZE_ERRORS
        self.sanitize_error_msg: str = SANITIZE_ERROR_MSG

        if static_assets is not None:
            if not os.path.isdir(static_assets):
                raise ValueError(f"static_assets must be a directory: {static_assets}")
            if not os.path.isabs(static_assets):
                raise ValueError(
                    f"static_assets must be an absolute path: {static_assets}"
                )

        self._static_assets: Union[str, os.PathLike[str], None] = static_assets

        self._sessions: Dict[str, Session] = {}

        self._sessions_needing_flush: Dict[int, Session] = {}

        self._registered_dependencies: Dict[str, HTMLDependency] = {}
        self._dependency_handler = starlette.routing.Router()

        if self._static_assets is not None:
            self._dependency_handler.routes.append(
                starlette.routing.Mount(
                    "/",
                    StaticFiles(directory=self._static_assets),
                    name="shiny-app-static-assets-directory",
                )
            )

        starlette_app = self.init_starlette_app()

        self.starlette_app = starlette_app

    def init_starlette_app(self):
        routes: list[starlette.routing.BaseRoute] = [
            starlette.routing.WebSocketRoute("/websocket/", self._on_connect_cb),
            starlette.routing.Route("/", self._on_root_request_cb, methods=["GET"]),
            starlette.routing.Route(
                "/session/{session_id}/{action}/{subpath:path}",
                self._on_session_request_cb,
                methods=["GET", "POST"],
            ),
            starlette.routing.Mount("/", app=self._dependency_handler),
        ]
        middleware: list[starlette.middleware.Middleware] = []
        if autoreload_url():
            shared_dir = os.path.join(os.path.dirname(__file__), "www", "shared")
            routes.insert(
                0,
                starlette.routing.Mount(
                    "/__shared",
                    app=StaticFiles(directory=shared_dir),
                ),
            )
            middleware.append(
                starlette.middleware.Middleware(InjectAutoreloadMiddleware)
            )
        # In Pyodide mode, an HTTPException(404) being thrown resulted in
        # some default error handler (that happened not to be async) being
        # run in a threadpool, which Pyodide could not handle. So in Pyodide
        # mode, install our own async error handler at the outermost layer
        # that we can.
        if is_pyodide:
            middleware.append(starlette.middleware.Middleware(ErrorMiddleware))

        starlette_app = starlette.applications.Starlette(
            routes=routes,
            middleware=middleware,
            lifespan=self._lifespan,
        )

        return starlette_app

    @contextlib.asynccontextmanager
    async def _lifespan(self, app: starlette.applications.Starlette):
        unreg = on_flushed(self._on_reactive_flushed, once=False)
        try:
            yield
        finally:
            unreg()

    async def _on_reactive_flushed(self):
        await self._flush_pending_sessions()

    def _create_session(self, conn: Connection) -> Session:
        id = secrets.token_hex(32)
        session = Session(self, id, conn, debug=self._debug)
        self._sessions[id] = session
        return session

    def _remove_session(self, session: Union[Session, str]) -> None:
        if isinstance(session, Session):
            session = session.id

        if self._debug:
            print(f"remove_session: {session}", flush=True)
        del self._sessions[session]

    def run(self, **kwargs: object) -> None:
        """
        Run the app.

        Parameters
        ----------
        kwargs
            Keyword arguments passed to :func:`~shiny.run_app`.
        """
        from ._main import run_app

        run_app(self, **kwargs)

    # ASGI entrypoint. Handles HTTP, WebSocket, and lifespan.
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.starlette_app(scope, receive, send)

    async def call_pyodide(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Communicate with pyodide.

        Warning
        -------
        This method is not intended for public usage. It's exported for use by
        shinylive.
        """

        # TODO: Pretty sure there are objects that need to be destroy()'d here?
        scope = cast(Any, scope).to_py()

        # ASGI requires some values to be byte strings, not character strings. Those are
        # not that easy to create in JavaScript, so we let the JS side pass us strings
        # and we convert them to bytes here.
        if "headers" in scope:
            # JS doesn't have `bytes` so we pass as strings and convert here
            scope["headers"] = [
                [value.encode("latin-1") for value in header]
                for header in scope["headers"]
            ]
        if "query_string" in scope and scope["query_string"]:
            scope["query_string"] = scope["query_string"].encode("latin-1")
        if "raw_path" in scope and scope["raw_path"]:
            scope["raw_path"] = scope["raw_path"].encode("latin-1")

        async def rcv() -> Message:
            event = await receive()
            return cast(Message, cast(Any, event).to_py())

        async def snd(event: Message):
            await send(event)

        await self(scope, rcv, snd)

    async def stop(self) -> None:
        """
        Stop the app (i.e., close all sessions).

        See Also
        --------
        ~shiny.Session.close
        """
        # convert to list to avoid modifying the dict while iterating over it, which
        # throws an error
        for session in list(self._sessions.values()):
            await session.close()

    # ==========================================================================
    # Connection callbacks
    # ==========================================================================
    async def _on_root_request_cb(self, request: Request) -> Response:
        """
        Callback passed to the ConnectionManager which is invoked when a HTTP
        request for / occurs.
        """
        self._ensure_web_dependencies(self.ui["dependencies"])
        return HTMLResponse(content=self.ui["html"])

    async def _on_connect_cb(self, ws: starlette.websockets.WebSocket) -> None:
        """
        Callback which is invoked when a new WebSocket connection is established.
        """
        await ws.accept()
        conn = StarletteConnection(ws)
        session = self._create_session(conn)

        await session._run()

    async def _on_session_request_cb(self, request: Request) -> ASGIApp:
        """
        Callback passed to the ConnectionManager which is invoked when a HTTP
        request for /session/* occurs.
        """
        session_id: str = request.path_params["session_id"]  # type: ignore
        action: str = request.path_params["action"]  # type: ignore
        subpath: str = request.path_params["subpath"]  # type: ignore

        if session_id in self._sessions:
            session: Session = self._sessions[session_id]
            with session_context(session):
                return await session._handle_request(request, action, subpath)

        return JSONResponse({"detail": "Not Found"}, status_code=404)

    # ==========================================================================
    # Flush
    # ==========================================================================
    def _request_flush(self, session: Session) -> None:
        # TODO: Until we have reactive domains, because we can't yet keep track
        # of which sessions need a flush.
        pass
        # self._sessions_needing_flush[session.id] = session

    async def _flush_pending_sessions(self) -> None:
        # TODO: Until we have reactive domains, flush all sessions (because we
        # can't yet keep track of which ones need a flush)
        # Use list() to create a copy, in case self._sessions mutates from under us,
        # which would cause a "dictionary changed size during iteration" RuntimeError
        for _, session in list(self._sessions.items()):
            await session._flush()
        # for id, session in self._sessions_needing_flush.items():
        #     await session.flush()
        #     del self._sessions_needing_flush[id]

    # ==========================================================================
    # HTML Dependency stuff
    # ==========================================================================
    def _ensure_web_dependencies(self, deps: List[HTMLDependency]) -> None:
        for dep in deps:
            self._register_web_dependency(dep)

    def _register_web_dependency(self, dep: HTMLDependency) -> None:
        if (
            dep.name in self._registered_dependencies
            and dep.version >= self._registered_dependencies[dep.name].version
        ):
            return

        # For HTMLDependencies that have sources on disk, mount the source dir.
        # (Some HTMLDependencies only carry head content, and have no source on disk.)
        if dep.source:
            paths = dep.source_path_map(lib_prefix=self.lib_prefix)
            self._dependency_handler.routes.insert(
                0,
                starlette.routing.Mount(
                    "/" + paths["href"],
                    StaticFiles(directory=paths["source"]),
                    name=dep.name + "-" + str(dep.version),
                ),
            )

        self._registered_dependencies[dep.name] = dep


def _render_page(ui: Union[Tag, TagList], lib_prefix: str) -> RenderedHTML:
    ui_res = copy.copy(ui)
    # Make sure requirejs, jQuery, and Shiny come before any other dependencies.
    # (see require_deps() for a comment about why we even include it)
    ui_res.insert(0, [require_deps(), jquery_deps(), shiny_deps()])
    return HTMLDocument(ui_res).render(lib_prefix=lib_prefix)



from shiny import *

app_ui = ui.page_fluid(ui.input_action_button("btn", "Press me!"))


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    @reactive.event(input.btn)
    def _():
        ui.insert_ui(
            ui.p("Number of clicks: ", input.btn()), selector="#btn", where="afterEnd"
        )


app = App(app_ui, server)



import asyncio
import os
import io
from datetime import date
from typing import Any

from shiny import *
from shiny.ui import div, p

import matplotlib.pyplot as plt
import numpy as np


def make_example(id: str, label: str, title: str, desc: str, extra: Any = None):
    return ui.column(
        4,
        div(
            {"class": "card mb-4"},
            div(title, class_="card-header"),
            div(
                {"class": "card-body"},
                p(desc, class_="card-text text-muted"),
                extra,
                ui.download_button(id, label, class_="btn-primary"),
            ),
        ),
    )


app_ui = ui.page_fluid(
    ui.row(
        make_example(
            "download1",
            label="Download CSV",
            title="Simple case",
            desc="Downloads a pre-existing file, using its existing name on disk.",
        ),
    ),
    ui.row(
        make_example(
            "download2",
            label="Download plot",
            title="Dynamic data generation",
            desc="Downloads a PNG that's generated on the fly.",
            extra=[
                ui.input_text("title", "Plot title", "Random scatter plot"),
                ui.input_slider(
                    "num_points", "Number of data points", min=1, max=100, value=50
                ),
            ],
        ),
    ),
    ui.row(
        make_example(
            "download3",
            "Download",
            "Dynamic filename",
            "Demonstrates that filenames can be generated on the fly (and use Unicode characters!).",
        ),
    ),
    ui.row(
        make_example(
            "download4",
            "Download",
            "Failed downloads",
            "Throws an error in the download handler, download should not succeed.",
        ),
    ),
    ui.row(
        make_example(
            "download5",
            "Download",
            "Undefined download",
            "This button doesn't have corresponding server code registered to it, download should result in 404 error",
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @session.download()
    def download1():
        """
        This is the simplest case. The implementation simply returns the name of a file.
        Note that the function name (`download1`) determines which download_button()
        corresponds to this function.
        """

        path = os.path.join(os.path.dirname(__file__), "mtcars.csv")
        return path

    @session.download(filename="image.png")
    def download2():
        """
        Another way to implement a file download is by yielding bytes; either all at
        once, like in this case, or by yielding multiple times. When using this
        approach, you should pass a filename argument to @session.download, which
        determines what the browser will name the downloaded file.
        """

        print(input.num_points())
        x = np.random.uniform(size=input.num_points())
        y = np.random.uniform(size=input.num_points())
        plt.figure()
        plt.scatter(x, y)
        plt.title(input.title())
        with io.BytesIO() as buf:
            plt.savefig(buf, format="png")
            yield buf.getvalue()

    @session.download(
        filename=lambda: f"新型-{date.today().isoformat()}-{np.random.randint(100,999)}.csv"
    )
    async def download3():
        await asyncio.sleep(0.25)
        yield "one,two,three\n"
        yield "新,1,2\n"
        yield "型,4,5\n"

    @session.download(id="download4", filename="failuretest.txt")
    async def _():
        yield "hello"
        raise Exception("This error was caused intentionally")


app = App(app_ui, server)



from contextlib import contextmanager
from contextvars import ContextVar, Token
import re
from typing import Pattern, Union


class ResolvedId(str):
    def __call__(self, id: "Id") -> "ResolvedId":
        if isinstance(id, ResolvedId):
            return id

        validate_id(id)

        if self == "":
            return ResolvedId(id)
        else:
            return ResolvedId(self + "-" + id)


Root: ResolvedId = ResolvedId("")


Id = Union[str, ResolvedId]


def resolve_id(id: Id) -> ResolvedId:
    curr_ns = _current_namespace.get()
    return curr_ns(id)


# \w is a large set for unicode patterns, that's fine; we mostly want to avoid some
# special characters like space, comma, period, and especially dash
re_valid_id: Pattern[str] = re.compile("^\\.?\\w+$")


def validate_id(id: str):
    if not re_valid_id.match(id):
        raise ValueError(
            f"The string '{id}' is not a valid id; only letters, numbers, and "
            "underscore are permitted"
        )


_current_namespace: ContextVar[ResolvedId] = ContextVar(
    "current_namespace", default=Root
)


@contextmanager
def namespace_context(id: Union[Id, None]):
    namespace = resolve_id(id) if id else Root
    token: Token[ResolvedId] = _current_namespace.set(namespace)
    try:
        yield
    finally:
        _current_namespace.reset(token)



"""Tests for `shiny.datastructures`."""

from shiny._datastructures import PriorityQueueFIFO


def test_priority_queue_fifo():
    q: PriorityQueueFIFO[str] = PriorityQueueFIFO()

    # The random-seeming items are here to ensure that the value of the items
    # do not affect the order that they go into the queue.
    q.put(1, "9")
    q.put(1, "8")
    q.put(2, "6")
    q.put(2, "7")

    assert q.get() == "6"
    assert q.get() == "7"
    assert q.get() == "9"
    assert q.get() == "8"



from shiny import *

app_ui = ui.page_fluid(
    ui.input_text("txt", "Enter the text to display below:"),
    ui.row(
        ui.column(6, ui.output_text("text")),
        ui.column(6, ui.output_text_verbatim("verb", placeholder=True)),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.text
    def text():
        return input.txt()

    @output
    @render.text
    def verb():
        return input.txt()


app = App(app_ui, server)



__all__ = ("input_numeric",)

from typing import Optional

from htmltools import tags, Tag, div, css, TagChildArg

from .._docstring import add_example
from .._namespaces import resolve_id
from ._utils import shiny_input_label


@add_example()
def input_numeric(
    id: str,
    label: TagChildArg,
    value: float,
    *,
    min: Optional[float] = None,
    max: Optional[float] = None,
    step: Optional[float] = None,
    width: Optional[str] = None,
) -> Tag:
    """
    Create an input control for entry of numeric values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    step
        Interval to use when stepping between min and max.
    width
        The CSS width, e.g. '400px', or '100%'

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        A numeric value.

    See Also
    -------
    ~shiny.ui.update_numeric
    """

    return div(
        shiny_input_label(id, label),
        tags.input(
            id=resolve_id(id),
            type="number",
            class_="form-control",
            value=value,
            min=min,
            max=max,
            step=step,
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )



from shiny import *
import numpy as np
import matplotlib.pyplot as plt

app_ui = ui.page_fluid(
    ui.input_slider("n", "Number of observations", min=0, max=1000, value=500),
    ui.input_action_button("go", "Go!", class_="btn-success"),
    ui.output_plot("plot"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.plot(alt="A histogram")
    def plot():
        # Take a reactive dependency on the action button...
        input.go()

        # ...but don't take a reactive dependency on the slider
        with reactive.isolate():
            np.random.seed(19680801)
            x = 100 + 15 * np.random.randn(input.n())

        fig, ax = plt.subplots()
        ax.hist(x, bins=30, density=True)
        return fig


app = App(app_ui, server)



from shiny import *
from datetime import date, timedelta

app_ui = ui.page_fluid(
    ui.input_slider("n", "Day of month", min=1, max=30, value=10),
    ui.input_date("inDate", "Input date"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        d = date(2013, 4, input.n())
        ui.update_date(
            "inDate",
            label="Date label " + str(input.n()),
            value=d,
            min=d - timedelta(days=3),
            max=d + timedelta(days=3),
        )


app = App(app_ui, server)



from typing import TypeVar, Generic, Tuple
from queue import PriorityQueue


T = TypeVar("T")


class PriorityQueueFIFO(Generic[T]):
    """
    Similar to queue.PriorityQueue, except that if two elements have the same
    priority, they are returned in the order they were inserted. Also, the item
    is kept separate from the priority value (with PriorityQueue, the priority
    is part of the item).
    """

    def __init__(self) -> None:
        # Using Tuple instead of tuple because in Python 3.8 and earlier, tuple isn't
        # generic
        self._pq: PriorityQueue[Tuple[int, int, T]] = PriorityQueue()
        self._counter: int = 0

    def put(self, priority: int, item: T) -> None:
        """
        Add an item to the queue.

        Parameters:
           priority (int): The priority of the item. Higher priority items will
                           come out of the queue before lower priority items.
           item (T): The item to put in the queue.
        """
        self._counter += 1
        self._pq.put((-priority, self._counter, item))

    def get(self) -> T:
        iteminfo: Tuple[int, int, T] = self._pq.get()
        return iteminfo[2]

    def empty(self) -> bool:
        return self._pq.empty()



import pytest
import random
from shiny._utils import AsyncCallbacks, Callbacks, private_seed
from typing import List


def test_randomness():

    current_state = random.getstate()
    try:
        # Make sure the public stream of randomness is independent of the private stream
        # https://github.com/rstudio/py-shiny/issues/140
        pub = random.randint(0, 100000000)
        with private_seed():
            priv = random.randint(0, 100000000)
        pub2 = random.randint(0, 100000000)
        with private_seed():
            priv2 = random.randint(0, 100000000)
        assert pub != priv and priv != pub2 and pub2 != priv2

        # By setting the same seed, we should get the same randomness
        random.seed(0)
        public = [random.randint(0, 100000000) for _ in range(3)]
        with private_seed():
            random.seed(0)
            private = [random.randint(0, 100000000) for _ in range(3)]
        assert public == private

        # Interleaved calls to private and public should give the same randomness
        public: List[float] = []
        private: List[float] = []
        random.seed(0)
        with private_seed():
            random.seed(0)
        public.append(random.randint(0, 100000000))
        with private_seed():
            private.append(random.randint(0, 100000000))
        with private_seed():
            private.append(random.randint(0, 100000000))
        public.append(random.randint(0, 100000000))
        public.append(random.randint(0, 100000000))
        with private_seed():
            private.append(random.randint(0, 100000000))
        assert public == private

    finally:
        random.setstate(current_state)


def test_callbacks():
    class MockCallback:
        def __init__(self):
            self.exec_count = 0

        def __call__(self):
            self.exec_count += 1

    callbacks = Callbacks()

    cb1 = MockCallback()
    cb2 = MockCallback()
    cb3 = MockCallback()
    cb4 = MockCallback()

    def mutate_registrations():
        # Unregister cb2
        cb2_reg_handle()
        # Calling registrations multiple times should just no-op
        cb2_reg_handle()

        callbacks.register(cb4)

    _ = callbacks.register(cb1)
    mut_reg_handle = callbacks.register(mutate_registrations)
    cb2_reg_handle = callbacks.register(cb2)
    _ = callbacks.register(cb3, once=True)

    callbacks.invoke()
    assert cb1.exec_count == 1  # Regular registration was called
    assert cb2.exec_count == 1  # Unregistered by an earlier handler, but still called
    assert cb3.exec_count == 1  # once=True registration was called
    assert cb4.exec_count == 0  # Registered during invoke(), not called

    mut_reg_handle()

    callbacks.invoke()
    assert cb1.exec_count == 2  # Regular registration was called again
    assert cb2.exec_count == 1  # Unregistered by previous invoke, not called again
    assert cb3.exec_count == 1  # once=True, so not called again
    assert cb4.exec_count == 1  # Registered during previous invoke(), was called


@pytest.mark.asyncio
async def test_async_callbacks():
    class AsyncMockCallback:
        def __init__(self):
            self.exec_count = 0

        async def __call__(self):
            self.exec_count += 1

    callbacks = AsyncCallbacks()

    cb1 = AsyncMockCallback()
    cb2 = AsyncMockCallback()
    cb3 = AsyncMockCallback()
    cb4 = AsyncMockCallback()

    async def mutate_registrations():
        # Unregister cb2
        cb2_reg_handle()
        # Calling registrations multiple times should just no-op
        cb2_reg_handle()

        callbacks.register(cb4)

    _ = callbacks.register(cb1)
    mut_reg_handle = callbacks.register(mutate_registrations)
    cb2_reg_handle = callbacks.register(cb2)
    _ = callbacks.register(cb3, once=True)

    await callbacks.invoke()
    assert cb1.exec_count == 1  # Regular registration was called
    assert cb2.exec_count == 1  # Unregistered by an earlier handler, but still called
    assert cb3.exec_count == 1  # once=True registration was called
    assert cb4.exec_count == 0  # Registered during invoke(), not called

    mut_reg_handle()

    await callbacks.invoke()
    assert cb1.exec_count == 2  # Regular registration was called again
    assert cb2.exec_count == 1  # Unregistered by previous invoke, not called again
    assert cb3.exec_count == 1  # once=True, so not called again
    assert cb4.exec_count == 1  # Registered during previous invoke(), was called



__all__ = (
    "page_navbar",
    "page_fluid",
    "page_fixed",
    "page_bootstrap",
)

import sys
from typing import Optional, Any, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from htmltools import (
    tags,
    Tag,
    TagList,
    div,
    TagChildArg,
)

# Tagifiable isn't used directly in this file, but it seems to necessary to import
# it somewhere for Sphinx to work cleanly.
from htmltools import Tagifiable  # pyright: ignore[reportUnusedImport] # noqa: F401

from .._docstring import add_example
from ._html_dependencies import bootstrap_deps
from ._navs import navset_bar
from .._namespaces import resolve_id
from ..types import MISSING, MISSING_TYPE, NavSetArg
from ._utils import get_window_title


def page_navbar(
    *args: NavSetArg,
    title: Optional[Union[str, Tag, TagList]] = None,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    position: Literal["static-top", "fixed-top", "fixed-bottom"] = "static-top",
    header: Optional[TagChildArg] = None,
    footer: Optional[TagChildArg] = None,
    bg: Optional[str] = None,
    inverse: bool = False,
    collapsible: bool = True,
    fluid: bool = True,
    window_title: Union[str, MISSING_TYPE] = MISSING,
    lang: Optional[str] = None
) -> Tag:
    """
    Create a navbar with a navs bar and a title.

    Parameters
    ----------

    args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the navigation elements into a menu when the
        width of the browser is less than 940 pixels (useful for viewing on smaller
        touchscreen device)
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.

    Returns
    -------
    A UI element.

    See Also
    -------
    :func:`~shiny.ui.nav`
    :func:`~shiny.ui.nav_menu`
    :func:`~shiny.ui.navset_bar`
    :func:`~shiny.ui.page_fluid`

    Example
    -------
    See :func:`~shiny.ui.nav`.
    """

    return tags.html(
        get_window_title(title, window_title),
        tags.body(
            navset_bar(
                *args,
                title=title,
                id=resolve_id(id) if id else None,
                selected=selected,
                position=position,
                header=header,
                footer=footer,
                bg=bg,
                inverse=inverse,
                collapsible=collapsible,
                fluid=fluid
            )
        ),
        lang=lang,
    )


@add_example()
def page_fluid(
    *args: Any, title: Optional[str] = None, lang: Optional[str] = None, **kwargs: str
) -> Tag:
    """
    Create a fluid page.

    Parameters
    ----------

    args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set as
        a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This will
        be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The default,
        `None`, results in an empty string.
    kwargs
        Attributes on the page level container.

    Returns
    -------
    A UI element.

    See Also
    -------
    :func:`~shiny.ui.page_fixed`
    :func:`~shiny.ui.page_bootstrap`
    :func:`~shiny.ui.page_navbar`
    """

    return page_bootstrap(
        div({"class": "container-fluid"}, *args, **kwargs), title=title, lang=lang
    )


@add_example()
def page_fixed(
    *args: Any, title: Optional[str] = None, lang: Optional[str] = None, **kwargs: str
) -> Tag:
    """
    Create a fixed page.

    Parameters
    ----------

    args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set as
        a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This will
        be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The default,
        `None`, results in an empty string.

    kwargs
        Attributes on the page level container.
    Returns
    -------
    A UI element.

    See Also
    -------
    :func:`~shiny.ui.page_fluid`
    :func:`~shiny.ui.page_bootstrap`
    :func:`~shiny.ui.page_navbar`
    """

    return page_bootstrap(
        div({"class": "container"}, *args, **kwargs), title=title, lang=lang
    )


# TODO: implement theme (just Bootswatch for now?)
def page_bootstrap(
    *args: Any, title: Optional[str] = None, lang: Optional[str] = None
) -> Tag:
    """
    Create a Bootstrap UI page container.

    Parameters
    ----------

    args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set as
        a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This will
        be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The default,
        `None`, results in an empty string.

    Returns
    -------
    A UI element.

    See Also
    -------
    :func:`~shiny.ui.page_fluid`
    :func:`~shiny.ui.page_navbar`
    """

    page = TagList(*bootstrap_deps(), *args)
    head = tags.title(title) if title else None
    return tags.html(tags.head(head), tags.body(page), lang=lang)



__all__ = ("Progress",)

from types import TracebackType
from typing import Optional, Type, Union
from warnings import warn

from .._docstring import add_example
from .._utils import rand_hex
from ..session import Session, require_active_session


@add_example()
class Progress:
    """
    Initialize a progress bar.

    Parameters
    ----------
    min
        The value that represents the starting point of the progress bar. Must be less
        than ``max``.
    max
        The value that represents the end of the progress bar. Must be greater than
        ``min``.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.
    """

    _style = "notification"

    min: int
    max: int
    value: Union[float, None]

    def __init__(
        self, min: int = 0, max: int = 1, session: Optional[Session] = None
    ) -> None:
        self.min = min
        self.max = max
        self.value = None
        self._id = rand_hex(8)
        self._closed = False
        self._session = require_active_session(session)

        msg = {"id": self._id, "style": self._style}
        self._session._send_progress("open", msg)

    def __enter__(self) -> "Progress":
        return self

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> bool:
        self.close()
        return False

    def set(
        self,
        value: Optional[float] = None,
        message: Optional[str] = None,
        detail: Optional[str] = None,
    ) -> None:
        """
        Updates the progress panel. When called the first time, the progress panel is
        displayed.

        Parameters
        ----------
        self
            The object instance
        value
            The value at which to set the progress bar, relative to ``min`` and ``max``.
            ``None`` hides the progress bar, if it is currently visible.
        message
            The message to be displayed to the user or ``None`` to hide the current
            message (if any).
        detail
            The detail message to be displayed to the user or ``None`` to hide the
            current detail message (if any). The detail message will be shown with a
            de-emphasized appearance relative to message.
        """

        if self._closed:
            warn("Attempting to set progress, but progress already closed.")
            return None

        self.value = value
        if value:
            # Normalize value to number between 0 and 1
            value = min(1, max(0, (value - self.min) / (self.max - self.min)))

        msg = {
            "id": self._id,
            "message": message,
            "detail": detail,
            "value": value,
            "style": self._style,
        }

        self._session._send_progress(
            "update", {k: v for k, v in msg.items() if v is not None}
        )

    def inc(
        self,
        amount: float = 0.1,
        message: Optional[str] = None,
        detail: Optional[str] = None,
    ) -> None:
        """
        Increment the progress bar.

        Like ``set``, this updates the progress panel. The difference is that ``inc``
        increases the progress bar by amount, instead of setting it to a specific value.

        Parameters
        ----------
        self
            The object instance
        amount
            The amount to increment in progress.
        message
            The message to be displayed to the user or ``None`` to hide the current
            message (if any).
        detail
            The detail message to be displayed to the user or ``None`` to hide the current
            detail message (if any). The detail message will be shown with a
            de-emphasized appearance relative to message.
        """

        if self.value is None:
            self.value = self.min

        value = min(self.value + amount, self.max)
        self.set(value, message, detail)

    def close(self) -> None:
        """
        Close the progress bar. You can also use the Progress object as a context
        manager, which will cause the progress bar to close on exit.

        Parameters
        ----------
        self
            The object instance

        Note
        ----
        Removes the progress panel. Future calls to set and close will be ignored.
        """
        if self._closed:
            warn("Attempting to close progress, but progress already closed.")
            return None

        self._session._send_progress("close", {"id": self._id, "style": self._style})
        self._closed = True



from ._core import (  # noqa: F401
    isolate,
    invalidate_later,
    flush,
    on_flushed,
    get_current_context,  # pyright: ignore[reportUnusedImport]
)
from ._poll import poll, file_reader
from ._reactives import (  # noqa: F401
    Value,
    Calc,
    Calc_,  # pyright: ignore[reportUnusedImport]
    CalcAsync_,  # pyright: ignore[reportUnusedImport]
    Effect,
    Effect_,  # pyright: ignore[reportUnusedImport]
    event,
)


__all__ = (
    "flush",
    "on_flushed",
    "Value",
    "Calc",
    "Effect",
    "event",
    "isolate",
    "invalidate_later",
    "poll",
    "file_reader",
)



"""
Tools for creating user interfaces including: custom components, HTML components,
layout helpers, page-level containers, and more.
"""

from ._bootstrap import (
    row,
    column,
    layout_sidebar,
    panel_well,
    panel_sidebar,
    panel_main,
    panel_conditional,
    panel_title,
    panel_fixed,
    panel_absolute,
    help_text,
)
from ._download_button import download_button, download_link
from ._input_action_button import input_action_button, input_action_link
from ._input_check_radio import (
    input_checkbox,
    input_checkbox_group,
    input_switch,
    input_radio_buttons,
)
from ._input_date import input_date, input_date_range
from ._input_file import input_file
from ._input_numeric import input_numeric
from ._input_password import input_password
from ._input_select import input_select, input_selectize
from ._input_slider import input_slider, SliderValueArg, SliderStepArg, AnimationOptions
from ._input_text import input_text, input_text_area
from ._input_update import (
    update_action_button,
    update_action_link,
    update_checkbox,
    update_switch,
    update_checkbox_group,
    update_radio_buttons,
    update_date,
    update_date_range,
    update_numeric,
    update_select,
    update_selectize,
    update_slider,
    update_text,
    update_text_area,
    update_navs,
)
from ._insert import insert_ui, remove_ui
from ._markdown import markdown
from ._modal import modal_button, modal, modal_show, modal_remove
from ._navs import (
    nav,
    nav_menu,
    nav_control,
    nav_spacer,
    navset_tab,
    navset_tab_card,
    navset_pill,
    navset_pill_card,
    navset_pill_list,
    navset_hidden,
    navset_bar,
)
from ._notification import notification_show, notification_remove
from ._output import (
    output_plot,
    output_image,
    output_text,
    output_text_verbatim,
    output_table,
    output_ui,
)
from ._page import page_navbar, page_fluid, page_fixed, page_bootstrap
from ._progress import Progress

from htmltools import (
    TagList,
    Tag,
    TagChildArg,
    TagAttrArg,
    tags,
    HTML,
    head_content,
    p,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    a,
    br,
    div,
    span,
    pre,
    code,
    img,
    strong,
    em,
    hr,
)


__all__ = (
    "row",
    "column",
    "layout_sidebar",
    "panel_well",
    "panel_sidebar",
    "panel_main",
    "panel_conditional",
    "panel_title",
    "panel_fixed",
    "panel_absolute",
    "help_text",
    "download_button",
    "download_link",
    "input_action_button",
    "input_action_link",
    "input_checkbox",
    "input_checkbox_group",
    "input_switch",
    "input_radio_buttons",
    "input_date",
    "input_date_range",
    "input_file",
    "input_numeric",
    "input_password",
    "input_select",
    "input_selectize",
    "input_slider",
    "SliderValueArg",
    "SliderStepArg",
    "AnimationOptions",
    "input_text",
    "input_text_area",
    "update_action_button",
    "update_action_link",
    "update_checkbox",
    "update_switch",
    "update_checkbox_group",
    "update_radio_buttons",
    "update_date",
    "update_date_range",
    "update_numeric",
    "update_select",
    "update_selectize",
    "update_slider",
    "update_text",
    "update_text_area",
    "update_navs",
    "insert_ui",
    "remove_ui",
    "markdown",
    "modal_button",
    "modal",
    "modal_show",
    "modal_remove",
    "nav",
    "nav_menu",
    "nav_control",
    "nav_spacer",
    "navset_tab",
    "navset_tab_card",
    "navset_pill",
    "navset_pill_card",
    "navset_pill_list",
    "navset_hidden",
    "navset_bar",
    "notification_show",
    "notification_remove",
    "output_plot",
    "output_image",
    "output_text",
    "output_text_verbatim",
    "output_table",
    "output_ui",
    "page_navbar",
    "page_fluid",
    "page_fixed",
    "page_bootstrap",
    "Progress",
    # Items below are from htmltools
    "TagList",
    "Tag",
    "TagChildArg",
    "TagAttrArg",
    "tags",
    "HTML",
    "head_content",
    "p",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "a",
    "br",
    "div",
    "span",
    "pre",
    "code",
    "img",
    "strong",
    "em",
    "hr",
)



from htmltools import HTMLDependency


def shiny_deps() -> HTMLDependency:
    return HTMLDependency(
        name="shiny",
        version="0.0.1",
        source={"package": "shiny", "subdir": "www/shared/"},
        script={"src": "shiny.js"},
        stylesheet={"href": "shiny.min.css"},
    )


def jquery_deps() -> HTMLDependency:
    return HTMLDependency(
        name="jquery",
        version="3.6.0",
        source={"package": "shiny", "subdir": "www/shared/jquery/"},
        script={"src": "jquery-3.6.0.min.js"},
    )


# Shiny doesn't (currently) use requirejs directly, but it does include it because a
# custom requirejs setup is need to get HTMLDependency()s (i.e., loading JS via <script>
# tags) to be usable. At the moment, we're just setting `window.define.amd=false` after
# loading requirejs so that the typical UMD pattern won't result in an anonymous
# define() error.
# https://requirejs.org/docs/errors.html#mismatch
# https://github.com/umdjs/umd
#
# Someday, we may want the same/similar thing in R, but this definitely more of an
# immediate issue for Python since many Jupyter extensions use requirejs.
def require_deps() -> HTMLDependency:
    return HTMLDependency(
        name="requirejs",
        version="2.3.6",
        source={"package": "shiny", "subdir": "www/shared/requirejs/"},
        script={"src": "require.min.js"},
    )



from shiny import *
from shiny.types import SafeException

app_ui = ui.page_fluid(
    ui.input_action_button("safe", "Throw a safe error"),
    ui.output_ui("safe"),
    ui.input_action_button("unsafe", "Throw an unsafe error"),
    ui.output_ui("unsafe"),
    ui.input_text(
        "txt",
        "Enter some text below, then remove it. Notice how the text is never fully removed.",
    ),
    ui.output_ui("txt_out"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Calc
    def safe_click():
        req(input.safe())
        return input.safe()

    @output
    @render.ui
    def safe():
        raise SafeException(f"You've clicked {str(safe_click())} times")

    @output
    @render.ui
    def unsafe():
        req(input.unsafe())
        raise Exception(f"Super secret number of clicks: {str(input.unsafe())}")

    @reactive.Effect
    def _():
        req(input.unsafe())
        print("unsafe clicks:", input.unsafe())
        # raise Exception("Observer exception: this should cause a crash")

    @output
    @render.ui
    def txt_out():
        req(input.txt(), cancel_output=True)
        return input.txt()


app = App(app_ui, server)
app.sanitize_errors = True



from typing import Dict, cast

import starlette.exceptions as exceptions
import starlette.responses as responses
from starlette.types import ASGIApp, Receive, Scope, Send


class ErrorMiddleware:
    """Inserts shiny-autoreload.js into the head.

    It's necessary to do it using middleware instead of in a nice htmldependency,
    because we want autoreload to be effective even when displaying an error page.
    """

    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        try:
            return await self.app(scope, receive, send)
        except exceptions.HTTPException as e:
            resp = responses.PlainTextResponse(
                e.detail,
                e.status_code,
                headers=cast(
                    Dict[str, str],
                    e.headers,  # pyright: ignore[reportUnknownMemberType]
                ),
                media_type="text/plain",
            )
            await resp(scope, receive, send)
        except Exception as e:
            # Seems super weird this is just going to stdout, should we use logger or
            # at least stderr?
            print("Unhandled error: " + str(e))
            resp = responses.PlainTextResponse(
                "An internal server error occurred", 500, media_type="text/plain"
            )
            await resp(scope, receive, send)



__all__ = ("input_text", "input_text_area")

import sys
from typing import Optional

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from htmltools import Tag, TagChildArg, css, div, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._utils import shiny_input_label


@add_example()
def input_text(
    id: str,
    label: TagChildArg,
    value: str = "",
    *,
    width: Optional[str] = None,
    placeholder: Optional[str] = None,
    autocomplete: Optional[str] = "off",
    spellcheck: Optional[Literal["true", "false"]] = None,
) -> Tag:
    """
    Create an input control for entry of text values

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    width
        The CSS width, e.g. '400px', or '100%'
    placeholder
        A hint as to what can be entered into the control.
    autocomplete
        Whether to enable browser autocompletion of the text input (default is None).
        If None, then it will use the browser's default behavior. Other possible values
        include "on", "off", "name", "username", and "email". See
        https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/autocomplete for
        more.
    spellcheck
        Whether to enable browser spell checking of the text input (default is None). If
        None, then it will use the browser's default behavior.

    Returns
    -------
    A UI element

    Notes
    ------
    .. admonition:: Server value

        A string containing the current text input. The default value is ``""`` unless
        ``value`` is provided.

    See Also
    -------
    ~shiny.ui.input_text_area
    """

    return div(
        shiny_input_label(id, label),
        tags.input(
            id=resolve_id(id),
            type="text",
            class_="form-control",
            value=value,
            placeholder=placeholder,
            autocomplete=autocomplete,
            spellcheck=spellcheck,
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )


@add_example()
def input_text_area(
    id: str,
    label: TagChildArg,
    value: str = "",
    *,
    width: Optional[str] = None,
    height: Optional[str] = None,
    cols: Optional[int] = None,
    rows: Optional[int] = None,
    placeholder: Optional[str] = None,
    resize: Optional[str] = None,
    autocomplete: Optional[str] = None,
    spellcheck: Optional[Literal["true", "false"]] = None,
) -> Tag:
    """
    Create a textarea input control for entry of unstructured text values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    width
        The CSS width, e.g. '400px', or '100%'
    height
        The CSS height, e.g. '400px', or '100%'
    cols
        Value of the visible character columns of the input, e.g. 80. This argument will
        only take effect if there is not a CSS width rule defined for this element; such
        a rule could come from the width argument of this function or from a containing
        page layout such as :func:`~shiny.ui.page_fluid`.
    rows
        The value of the visible character rows of the input, e.g. 6. If the height
        argument is specified, height will take precedence in the browser's rendering.
    placeholder
        A hint as to what can be entered into the control.
    resize
        Which directions the textarea box can be resized. Can be one of "both", "none",
        "vertical", and "horizontal". The default, ``None``, will use the client
        browser's default setting for resizing textareas.
    autocomplete
        Whether to enable browser autocompletion of the text input (default is "off").
        If None, then it will use the browser's default behavior. Other possible values
        include "on", "name", "username", and "email". See
        https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/autocomplete for
        more.
    spellcheck
        Whether to enable browser spell checking of the text input (default is None). If
        None, then it will use the browser's default behavior.

    Returns
    -------
    A UI element

    Notes
    ------
    .. admonition:: Server value

        A string containing the current text input. The default value is ``""`` unless
        ``value`` is provided.

    See Also
    -------
    ~shiny.ui.input_text
    """

    if resize and resize not in ["none", "both", "horizontal", "vertical"]:
        raise ValueError("Invalid resize value: " + str(resize))

    area = tags.textarea(
        value,
        id=resolve_id(id),
        class_="form-control",
        style=css(width=None if width else "100%", height=height, resize=resize),
        placeholder=placeholder,
        rows=rows,
        cols=cols,
        autocomplete=autocomplete,
        spellcheck=spellcheck,
    )

    return div(
        shiny_input_label(id, label),
        area,
        class_="form-group shiny-input-container",
        style=css(width=width),
    )



import pathlib

import pandas as pd
from shiny import *

dir = pathlib.Path(__file__).parent

app_ui = ui.page_fluid(ui.output_table("result"), class_="p-3")


@reactive.file_reader(dir / "mtcars.csv")
def read_file():
    return pd.read_csv(dir / "mtcars.csv")


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.table
    def result():
        return read_file()


app = App(app_ui, server)



import base64
import json
import os
import re
import shutil
import sys
from pathlib import Path
from typing import Callable, List, Optional, Union


if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

_SHINYLIVE_DOWNLOAD_URL = "https://pyshiny.netlify.app/shinylive"
_SHINYLIVE_DEFAULT_VERSION = "0.0.1dev"

# This is the same as the FileContentJson type in TypeScript.
class FileContentJson(TypedDict):
    name: str
    content: str
    type: Literal["text", "binary"]


def deploy_static(
    appdir: Union[str, Path],
    destdir: Union[str, Path],
    *,
    overwrite: bool = False,
    subdir: Union[str, Path, None] = None,
    version: str = _SHINYLIVE_DEFAULT_VERSION,
    verbose: bool = False,
) -> None:
    """
    Create a statically deployable distribution with a Shiny app.
    """

    def verbose_print(*args: object) -> None:
        if verbose:
            print(*args)

    appdir = Path(appdir)
    destdir = Path(destdir)

    if not (appdir / "app.py").exists():
        raise ValueError(f"Directory {appdir} must contain a file named app.py.")

    if subdir is None:
        subdir = ""
    subdir = Path(subdir)
    if subdir.is_absolute():
        raise ValueError("subdir must be a relative path")

    shinylive_bundle_dir = _ensure_shinylive_local(version=version)

    print(f"Copying {shinylive_bundle_dir}/ to {destdir}/")
    if not destdir.exists():
        destdir.mkdir()

    # destdir_rel: Union[str, None] = None
    # if _is_relative_to(destdir, appdir):
    #     destdir_rel = str(Path(destdir).relative_to(Path(appdir)))

    # After we drop Python 3.7 support, this can be replaced with the shutil.copytree
    # call below.
    _copy_recursive(
        shinylive_bundle_dir,
        destdir,
        copy_function=_copy_fn(overwrite, verbose_print=verbose_print),
    )
    # shutil.copytree(
    #     shinylive_bundle_dir,
    #     destdir,
    #     copy_function=_copy_fn(overwrite, verbose_print=verbose_print),
    #     dirs_exist_ok=True,
    # )

    app_files: List[FileContentJson] = []
    # Recursively iterate over files in app directory, and collect the files into
    # app_files data structure.
    exclude_names = {"__pycache__", "venv", ".venv"}
    for root, dirs, files in os.walk(appdir, topdown=True):
        root = Path(root)

        if _is_relative_to(Path(root), destdir):
            # In case destdir is inside of the appdir, don't copy those files.
            continue

        dirs[:] = [d for d in dirs if not d.startswith(".")]
        dirs[:] = set(dirs) - exclude_names
        rel_dir = root.relative_to(appdir)
        files = [f for f in files if not f.startswith(".")]
        files = [f for f in files if f not in exclude_names]
        files.sort()

        # Move app.py to first in list.
        if "app.py" in files:
            app_py_idx = files.index("app.py")
            files.insert(0, files.pop(app_py_idx))

        # Add the file to the app_files list.
        for filename in files:
            if rel_dir == ".":
                output_filename = filename
            else:
                output_filename = str(rel_dir / filename)

            if filename == "shinylive.js":
                print(
                    f"Warning: Found shinylive.js in source directory '{appdir}/{rel_dir}'. Are you including a shinylive distribution in your app?"
                )

            type: Literal["text", "binary"] = "text"
            try:
                with open(root / filename, "r") as f:
                    file_content = f.read()
                    type = "text"
            except UnicodeDecodeError:
                # If text failed, try binary.
                with open(root / filename, "rb") as f:
                    file_content_bin = f.read()
                    file_content = base64.b64encode(file_content_bin).decode("utf-8")
                    type = "binary"

            app_files.append(
                {
                    "name": output_filename,
                    "content": file_content,
                    "type": type,
                }
            )

    # Write the index.html, editor/index.html, and app.json in the destdir.
    html_source_dir = shinylive_bundle_dir / "shinylive/shiny_static"
    app_destdir = destdir / subdir

    # For a subdir like a/b/c, this will be ../../../
    subdir_inverse = "/".join([".."] * _path_length(subdir))
    if subdir_inverse != "":
        subdir_inverse += "/"

    if not app_destdir.exists():
        app_destdir.mkdir()

    _copy_file_and_substitute(
        src=html_source_dir / "index.html",
        dest=app_destdir / "index.html",
        search_str="{{REL_PATH}}",
        replace_str=subdir_inverse,
    )

    editor_destdir = app_destdir / "edit"
    if not editor_destdir.exists():
        editor_destdir.mkdir()
    _copy_file_and_substitute(
        src=html_source_dir / "edit" / "index.html",
        dest=(editor_destdir / "index.html"),
        search_str="{{REL_PATH}}",
        replace_str=subdir_inverse,
    )

    app_json_output_file = app_destdir / "app.json"

    print("Writing to " + str(app_json_output_file), end="")
    json.dump(app_files, open(app_json_output_file, "w"))
    print(":", app_json_output_file.stat().st_size, "bytes")

    print(
        f"\nRun the following to serve the app:\n  python3 -m http.server --directory {destdir} 8008"
    )


def _copy_fn(
    overwrite: bool, verbose_print: Callable[..., None] = lambda x: None
) -> Callable[..., None]:
    """Returns a function that can be used as a copy_function for shutil.copytree.

    If overwrite is True, the copy function will overwrite files that already exist.
    If overwrite is False, the copy function will not overwrite files that already exist.
    """

    def mycopy(src: str, dst: str, **kwargs: object) -> None:
        if os.path.exists(dst):
            if overwrite:
                verbose_print(f"Overwriting {dst}")
                os.remove(dst)
            else:
                verbose_print(f"Skipping {dst}")
                return

        shutil.copy2(src, dst, **kwargs)

    return mycopy


def _path_length(path: Union[str, Path]) -> int:
    """Returns the number of elements in a path.

    For example 'a' has length 1, 'a/b' has length 2, etc.
    """

    path = str(path)
    if os.path.isabs(path):
        raise ValueError("path must be a relative path")

    # Unfortunately, there's no equivalent of os.path.normpath for Path objects.
    path = os.path.normpath(path)
    if path == ".":
        return 0

    # On Windows, replace backslashes with forward slashes.
    if os.name == "nt":
        path.replace("\\", "/")

    return len(path.split("/"))


def _copy_file_and_substitute(
    src: Union[str, Path], dest: Union[str, Path], search_str: str, replace_str: str
) -> None:
    with open(src, "r") as fin:
        in_content = fin.read()
        in_content = in_content.replace(search_str, replace_str)
        with open(dest, "w") as fout:
            fout.write(in_content)


def remove_shinylive_local(
    shinylive_dir: Union[str, Path, None] = None,
    version: Optional[str] = _SHINYLIVE_DEFAULT_VERSION,
) -> None:
    """Removes local copy of shinylive.

    Parameters
    ----------
    shinylive_dir
        The directory where shinylive is stored. If None, the default directory will
        be used.

    version
        If a version is specified, only that version will be removed.
        If None, all local versions of shinylive will be removed.
    """

    if shinylive_dir is None:
        shinylive_dir = get_default_shinylive_dir()

    shinylive_dir = Path(shinylive_dir)

    target_dir = shinylive_dir
    if version is not None:
        target_dir = target_dir / f"shinylive-{version}"

    shutil.rmtree(target_dir)


def _ensure_shinylive_local(
    destdir: Union[Path, None] = None,
    version: str = _SHINYLIVE_DEFAULT_VERSION,
    url: str = _SHINYLIVE_DOWNLOAD_URL,
) -> Path:
    """Ensure that there is a local copy of shinylive."""

    if destdir is None:
        destdir = Path(get_default_shinylive_dir())

    if not destdir.exists():
        print("Creating directory " + str(destdir))
        destdir.mkdir()

    shinylive_bundle_dir = destdir / f"shinylive-{version}"
    if not shinylive_bundle_dir.exists():
        print(f"{shinylive_bundle_dir} does not exist.")
        download_shinylive(url=url, version=version, destdir=destdir)

    return shinylive_bundle_dir


def download_shinylive(
    destdir: Union[str, Path, None] = None,
    version: str = _SHINYLIVE_DEFAULT_VERSION,
    url: str = _SHINYLIVE_DOWNLOAD_URL,
) -> None:
    import tarfile
    import tempfile
    import urllib.request

    if destdir is None:
        destdir = get_default_shinylive_dir()

    destdir = Path(destdir)

    with tempfile.NamedTemporaryFile() as tmp:

        bundle_url = f"{url}/shinylive-{version}.tar.gz"
        print(f"Downloading {bundle_url}...")
        urllib.request.urlretrieve(bundle_url, tmp.name)

        print(f"Unzipping to {destdir}")
        with tarfile.open(tmp.name) as tar:
            tar.extractall(destdir)


def get_default_shinylive_dir() -> str:
    import appdirs

    return os.path.join(appdirs.user_cache_dir("shiny"), "shinylive")


def copy_shinylive_local(
    source_dir: Union[str, Path],
    destdir: Optional[Union[str, Path]] = None,
    version: str = _SHINYLIVE_DEFAULT_VERSION,
):
    if destdir is None:
        destdir = Path(get_default_shinylive_dir())

    destdir = Path(destdir)

    target_dir = destdir / ("shinylive-" + version)

    if target_dir.exists():
        shutil.rmtree(target_dir)

    shutil.copytree(source_dir, target_dir)


def _installed_shinylive_versions(shinylive_dir: Optional[Path] = None) -> List[str]:
    if shinylive_dir is None:
        shinylive_dir = Path(get_default_shinylive_dir())

    shinylive_dir = Path(shinylive_dir)
    subdirs = shinylive_dir.iterdir()
    subdirs = [re.sub("^shinylive-", "", str(s)) for s in subdirs]
    return subdirs


def print_shinylive_local_info() -> None:

    print(
        f"""    Local shinylive dir:
        {get_default_shinylive_dir()}

    Installed versions:
        {", ".join(_installed_shinylive_versions())}"""
    )


# A wrapper for shutil.copytree. In Python >= 3.8, we can use dirs_exist_ok=True so that
# it doesn't error if the destination dir or subdirs already exist. In <= 3.7, we need
# to ignore directories manually. After we drop 3.7 support, we can remove this wrapper
# function.
def _copy_recursive(
    source: Path,
    dest: Path,
    copy_function: Callable[..., None] = shutil.copy2,
) -> None:
    if sys.version_info >= (3, 8):
        shutil.copytree(source, dest, copy_function=copy_function, dirs_exist_ok=True)
    else:
        source = Path(source)
        dest = Path(dest)

        if source.is_dir():
            if not dest.is_dir():
                dest.mkdir()
            files = source.iterdir()
            for f in files:
                _copy_recursive(source / f, dest / f)
        else:
            shutil.copyfile(source, dest)


def _is_relative_to(path: Path, base: Path) -> bool:
    """
    Wrapper for `PurePath.is_relative_to`, which was added in Python 3.9.
    """
    if sys.version_info >= (3, 9):
        return path.is_relative_to(base)
    else:
        try:
            path.relative_to(base)
            return True
        except ValueError:
            return False



from typing import List

from htmltools import HTML, HTMLDependency

from ..html_dependencies import jquery_deps


def bootstrap_deps() -> List[HTMLDependency]:
    dep = HTMLDependency(
        name="bootstrap",
        version="5.0.1",
        source={"package": "shiny", "subdir": "www/shared/bootstrap/"},
        script={"src": "bootstrap.bundle.min.js"},
        stylesheet={"href": "bootstrap.min.css"},
    )
    deps = [jquery_deps(), dep]
    return deps


def ionrangeslider_deps() -> List[HTMLDependency]:
    return [
        HTMLDependency(
            name="ionrangeslider",
            version="2.3.1",
            source={"package": "shiny", "subdir": "www/shared/ionrangeslider/"},
            script={"src": "js/ion.rangeSlider.min.js"},
            stylesheet={"href": "css/ion.rangeSlider.css"},
        ),
        HTMLDependency(
            name="strftime",
            version="0.9.2",
            source={"package": "shiny", "subdir": "www/shared/strftime/"},
            script={"src": "strftime-min.js"},
        ),
    ]


def datepicker_deps() -> HTMLDependency:
    return HTMLDependency(
        name="bootstrap-datepicker",
        version="1.9.0",
        source={"package": "shiny", "subdir": "www/shared/datepicker/"},
        # TODO: pre-compile the Bootstrap 5 version?
        stylesheet={"href": "css/bootstrap-datepicker3.min.css"},
        script={"src": "js/bootstrap-datepicker.min.js"},
        # Need to enable noConflict mode. See #1346.
        head=HTML(
            "<script>(function() { var datepicker = $.fn.datepicker.noConflict(); $.fn.bsDatepicker = datepicker; })();</script>"
        ),
    )


def selectize_deps() -> HTMLDependency:
    return HTMLDependency(
        name="selectize",
        version="0.12.6",
        source={"package": "shiny", "subdir": "www/shared/selectize/"},
        script=[
            {"src": "js/selectize.min.js"},
            {"src": "accessibility/js/selectize-plugin-a11y.min.js"},
        ],
        # TODO: pre-compile the Bootstrap 5 version?
        stylesheet={"href": "css/selectize.bootstrap3.css"},
    )


def jqui_deps() -> HTMLDependency:
    return HTMLDependency(
        name="jquery-ui",
        version="1.12.1",
        source={"package": "shiny", "subdir": "www/shared/jqueryui/"},
        script={"src": "jquery-ui.min.js"},
        stylesheet={"href": "jquery-ui.min.css"},
    )



"""Reactive components"""

__all__ = ("Value", "Calc", "Calc_", "CalcAsync_", "Effect", "Effect_", "event")

import functools
import traceback
import warnings
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

from .. import _utils
from .._docstring import add_example
from .._utils import is_async_callable, run_coro_sync
from .._validation import req
from ..render import RenderFunction
from ..types import MISSING, MISSING_TYPE, ActionButtonValue, SilentException
from ._core import Context, Dependents, ReactiveWarning, isolate

if TYPE_CHECKING:
    from ..session import Session

T = TypeVar("T")

# ==============================================================================
# Value
# ==============================================================================
@add_example()
class Value(Generic[T]):
    """
    Create a reactive value

    Parameters
    ----------
    value
        An optional initial value.
    read_only
        If ``True``, then the reactive value cannot be `set()`.

    Returns
    -------
    An instance of a reactive value.

    Raises
    ------
    ~shiny.types.SilentException
        If :func:`~Value.get` is called before a value is provided/set.

    Note
    ----
    A reactive value may only be read from within a reactive function (e.g.,
    :func:`Calc`, :func:`Effect`, :func:`shiny.render.text`, etc.) and, when doing so,
    the function takes a reactive dependency on the value (i.e., when the value changes,
    the calling reactive function will re-execute).

    See Also
    --------
    ~shiny.Inputs
    Calc
    Effect
    """

    # These overloads are necessary so that the following hold:
    # - Value() is marked by the type checker as an error, because the type T is
    #   unknown. (It is not a run-time error.)
    # - Value[int]() works.
    # - Value[int](1) works.
    # - Value(1) works, with T is inferred to be int.
    @overload
    def __init__(
        self, value: MISSING_TYPE = MISSING, *, read_only: bool = False
    ) -> None:
        ...

    @overload
    def __init__(self, value: T, *, read_only: bool = False) -> None:
        ...

    # If `value` is MISSING, then `get()` will raise a SilentException, until a new
    # value is set. Calling `unset()` will set the value to MISSING.
    def __init__(
        self, value: Union[T, MISSING_TYPE] = MISSING, *, read_only: bool = False
    ) -> None:
        self._value: T = value
        self._read_only: bool = read_only
        self._value_dependents: Dependents = Dependents()
        self._is_set_dependents: Dependents = Dependents()

    def __call__(self) -> T:
        return self.get()

    def get(self) -> T:
        """
        Read the reactive value.

        Returns
        -------
        The reactive value.

        Raises
        ------
        ~shiny.types.SilentException
            If the value is not set.
        RuntimeError
            If called from outside a reactive function.
        """

        self._value_dependents.register()

        if isinstance(self._value, MISSING_TYPE):
            raise SilentException

        return self._value

    def set(self, value: T) -> bool:
        """
        Set the reactive value to a new value.

        Parameters
        ----------
        value
            A value.

        Returns
        -------
        ``True`` if the value was set to a different value and ``False`` otherwise.

        Raises
        ------
        RuntimeError
            If called on a read-only reactive value.
        """
        if self._read_only:
            raise RuntimeError(
                "Can't set read-only Value. If you are trying to set an input value, use `update_xxx()` instead."
            )
        return self._set(value)

    # The ._set() method allows setting read-only Value objects. This is used when the
    # Value is part of a session.Inputs object, and the session wants to set it.
    def _set(self, value: T) -> bool:
        if self._value is value:
            return False

        if isinstance(self._value, MISSING_TYPE) != isinstance(value, MISSING_TYPE):
            self._is_set_dependents.invalidate()

        self._value = value
        self._value_dependents.invalidate()
        return True

    def unset(self) -> None:
        """
        Unset the reactive value.

        Returns
        -------
        ``True`` if the value was set prior to this unsetting.
        """
        self.set(MISSING)  # type: ignore

    def is_set(self) -> bool:
        """
        Check if the reactive value is set.

        Returns
        -------
        ``True`` if the value is set, ``False`` otherwise.
        """

        self._is_set_dependents.register()
        return not isinstance(self._value, MISSING_TYPE)

    def freeze(self) -> None:
        """
        Freeze the reactive value.

        Freezing is equivalent to unsetting the value, but it does not invalidate
        dependents.
        """
        self._value = MISSING


# ==============================================================================
# Calc
# ==============================================================================

CalcFunction = Callable[[], T]
CalcFunctionAsync = Callable[[], Awaitable[T]]


class Calc_(Generic[T]):
    """
    Mark a function as a reactive calculation.

    Warning
    -------
    Most users shouldn't use this class directly to initialize a reactive calculation
    (instead, use the :func:`Calc` decorator).
    """

    def __init__(
        self,
        fn: CalcFunction[T],
        *,
        session: Union[MISSING_TYPE, "Session", None] = MISSING,
    ) -> None:
        self.__name__ = fn.__name__
        self.__doc__ = fn.__doc__

        # The CalcAsync subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: CalcFunctionAsync[T] = _utils.wrap_async(fn)
        self._is_async: bool = _utils.is_async_callable(fn)

        self._dependents: Dependents = Dependents()
        self._invalidated: bool = True
        self._running: bool = False
        self._most_recent_ctx_id: int = -1
        self._ctx: Optional[Context] = None
        self._exec_count: int = 0

        self._session: Optional[Session]
        # Use `isinstance(x, MISSING_TYPE)`` instead of `x is MISSING` because
        # the type checker doesn't know that MISSING is the only instance of
        # MISSING_TYPE; this saves us from casting later on.
        if isinstance(session, MISSING_TYPE):
            from ..session import get_current_session

            # If no session is provided, autodetect the current session (this
            # could be None if outside of a session).
            session = get_current_session()
        self._session = session

        # Use lists to hold (optional) value and error, instead of Optional[T],
        # because it makes typing more straightforward. For example if
        # .get_value() simply returned self._value, self._value had type
        # Optional[T], then the return type for get_value() would have to be
        # Optional[T].
        self._value: list[T] = []
        self._error: list[Exception] = []

    def __call__(self) -> T:
        # Run the Coroutine (synchronously), and then return the value.
        # If the Coroutine yields control, then an error will be raised.
        return _utils.run_coro_sync(self.get_value())

    # TODO: should this be private?
    async def get_value(self) -> T:
        self._dependents.register()

        if self._invalidated or self._running:
            await self.update_value()

        if self._error:
            raise self._error[0]

        return self._value[0]

    # TODO: should this be private?
    async def update_value(self) -> None:
        self._ctx = Context()
        self._most_recent_ctx_id = self._ctx.id

        self._ctx.on_invalidate(self._on_invalidate_cb)

        self._exec_count += 1
        self._invalidated = False

        was_running = self._running
        self._running = True

        from ..session import session_context

        with session_context(self._session):
            try:
                with self._ctx():
                    await self._run_func()
            finally:
                self._running = was_running

    def _on_invalidate_cb(self) -> None:
        self._invalidated = True
        self._value.clear()  # Allow old value to be GC'd
        self._dependents.invalidate()
        self._ctx = None  # Allow context to be GC'd

    async def _run_func(self) -> None:
        self._error.clear()
        try:
            self._value.append(await self._fn())
        except Exception as err:
            self._error.append(err)


class CalcAsync_(Calc_[T]):
    """
    Mark an async function as a reactive calculation.

    Warning
    -------
    Most users shouldn't use this class directly to initialize a reactive calculation
    (instead, use the :func:`Calc` decorator).
    """

    def __init__(
        self,
        fn: CalcFunctionAsync[T],
        *,
        session: Union[MISSING_TYPE, "Session", None] = MISSING,
    ) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")

        super().__init__(cast(CalcFunction[T], fn), session=session)

    async def __call__(self) -> T:  # pyright: ignore[reportIncompatibleMethodOverride]
        return await self.get_value()


@overload
def Calc(fn: CalcFunctionAsync[T]) -> CalcAsync_[T]:
    ...


@overload
def Calc(fn: CalcFunction[T]) -> Calc_[T]:
    ...


# Note that the specified return type of this Calc() overload (with a `session`) isn't
# exactly the same as the actual returned object -- the specified return type is a
# Callable that takes a CalcFunction[T], and actual return type is a Callable that takes
# CalcFunction[T] | CalcFunctionAsync[T]. Both are technically correct, since the
# CalcFunction's T encompasses both "regular" types V as well as Awatiable[V]. (We're
# using V to represent a generic type that is NOT itself Awaitable.) So if the T
# represents an Awaitable[V], then the type checker knows that the returned function
# will return a Calc[Awaitable[V]].
#
# However, if the Calc() function is specified to return a Callable that takes
# CalcFunction[T] | CalcFunctionAsync[T], then if a CalcFunctionAsync is passed in, the
# type check will not know that the returned Calc object is a Calc[Awaitable[V]]. It
# will think that it's a [Calc[V]]. Then the type checker will think that the returned
# Calc object is not async when it actually is.
#
# To work around this, we say that Calc() returns a Callable that takes a
# CalcFunction[T], instead of the union type. We're sort of tricking the type checker
# twice: once here, and once when we return a Calc object (which has a synchronous
# __call__ method) or CalcAsync object (which has an async __call__ method), and it
# works out.
@overload
def Calc(
    *, session: Union[MISSING_TYPE, "Session", None] = MISSING
) -> Callable[[CalcFunction[T]], Calc_[T]]:
    ...


@add_example()
def Calc(
    fn: Optional[Union[CalcFunction[T], CalcFunctionAsync[T]]] = None,
    *,
    session: Union[MISSING_TYPE, "Session", None] = MISSING,
) -> Union[Calc_[T], Callable[[CalcFunction[T]], Calc_[T]]]:
    """
    Mark a function as a reactive calculation.

    A reactive calculation is a function whose return value depends solely on other
    reactive value(s) (i.e., :class:`~shiny.Inputs`, :class:`~shiny.reactive.Value`,
    and other reactive calculations). Whenever a reactive value changes, any reactive
    calculations that depend on it are "invalidated" and automatically re-execute when
    necessary. If a reactive calculation is marked as invalidated, any other reactive
    calculations that recently called it are also marked as invalidated. In this way,
    invalidations ripple through reactive calculations that depend on each other.

    Parameters
    ----------
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    Returns
    -------
    A decorator that marks a function as a reactive calculation.

    Tip
    ---
    Reactive calculations should not produce any side effects; to reactively produce
    side effects, use :func:`~shiny.reactive.Effect` instead.

    See Also
    --------
    ~shiny.Inputs
    ~shiny.reactive.Value
    ~shiny.reactive.Effect
    ~shiny.reactive.invalidate_later
    ~shiny.event
    """

    def create_calc(fn: Union[CalcFunctionAsync[T], CalcFunction[T]]) -> Calc_[T]:
        if _utils.is_async_callable(fn):
            return CalcAsync_(fn, session=session)
        else:
            fn = cast(CalcFunction[T], fn)
            return Calc_(fn, session=session)

    if fn is None:
        return create_calc
    else:
        return create_calc(fn)


# ==============================================================================
# Effect
# ==============================================================================

EffectFunction = Callable[[], None]
EffectFunctionAsync = Callable[[], Awaitable[None]]


class Effect_:
    """
    Mark a function as a reactive side effect.

    Warning
    -------
    Most users shouldn't use this class directly to initialize a reactive side effect
    (instead, use the :func:`Effect` decorator).
    """

    def __init__(
        self,
        fn: Union[EffectFunction, EffectFunctionAsync],
        *,
        suspended: bool = False,
        priority: int = 0,
        session: Union[MISSING_TYPE, "Session", None] = MISSING,
    ) -> None:
        self.__name__ = fn.__name__
        self.__doc__ = fn.__doc__

        # The EffectAsync subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: EffectFunctionAsync = _utils.wrap_async(fn)
        # This indicates whether the user's effect function (before wrapping) is async.
        self._is_async: bool = _utils.is_async_callable(fn)

        self._priority: int = priority
        self._suspended = suspended
        self._on_resume: Callable[[], None] = lambda: None

        self._invalidate_callbacks: list[Callable[[], None]] = []
        self._destroyed: bool = False
        self._ctx: Optional[Context] = None
        self._exec_count: int = 0

        self._session: Optional[Session]
        # Use `isinstance(x, MISSING_TYPE)`` instead of `x is MISSING` because
        # the type checker doesn't know that MISSING is the only instance of
        # MISSING_TYPE; this saves us from casting later on.
        if isinstance(session, MISSING_TYPE):
            from ..session import get_current_session

            # If no session is provided, autodetect the current session (this
            # could be None if outside of a session).
            session = get_current_session()
        self._session = session

        if self._session is not None:
            self._session.on_ended(self._on_session_ended_cb)

        # Defer the first running of this until flushReact is called
        self._create_context().invalidate()

    def _create_context(self) -> Context:
        ctx = Context()

        # Store the context explicitly in Effect object
        # TODO: More explanation here
        self._ctx = ctx

        def on_invalidate_cb() -> None:
            # Context is invalidated, so we don't need to store a reference to it
            # anymore.
            self._ctx = None

            for cb in self._invalidate_callbacks:
                cb()

            if self._destroyed:
                return

            def _continue() -> None:
                ctx.add_pending_flush(self._priority)
                if self._session:
                    self._session._send_message_sync({"busy": "busy"})

            if self._suspended:
                self._on_resume = _continue
            else:
                _continue()

        async def on_flush_cb() -> None:
            if not self._destroyed:
                await self._run()
            if self._session:
                self._session._send_message_sync({"busy": "idle"})

        ctx.on_invalidate(on_invalidate_cb)
        ctx.on_flush(on_flush_cb)

        return ctx

    async def _run(self) -> None:
        ctx = self._create_context()
        self._exec_count += 1

        from ..session import session_context

        with session_context(self._session):
            try:
                with ctx():
                    await self._fn()
            except SilentException:
                # It's OK for SilentException to cause an Effect to stop running
                pass
            except Exception as e:
                traceback.print_exc()

                warnings.warn("Error in Effect: " + str(e), ReactiveWarning)
                if self._session:
                    await self._session._unhandled_error(e)

    def on_invalidate(self, callback: Callable[[], None]) -> None:
        """
        Register a callback that will be called when this reactive effect is
        invalidated.

        Parameters
        ----------
        callback
            A callback that will be called when this reactive effect is invalidated.
        """
        self._invalidate_callbacks.append(callback)

    def destroy(self) -> None:
        """
        Destroy this reactive effect.

        Stops the observer from executing ever again, even if it is currently scheduled
        for re-execution.
        """
        self._destroyed = True

        if self._ctx is not None:
            self._ctx.invalidate()

    def suspend(self) -> None:
        """
        Suspend the effect.

        Pauses scheduling of flushes (re-executions) in response to invalidations. If
        the effect was invalidated prior to this call but it has not re-executed yet
        (because it waits until on_flush is called) then that re-execution will still
        occur, because the flush is already scheduled.
        """
        self._suspended = True

    def resume(self) -> None:
        """
        Resume the effect.

        Causes this effect to start re-executing in response to invalidations. If the
        effect was invalidated while suspended, then it will schedule itself for
        re-execution (pending flush).
        """
        if self._suspended:
            self._suspended = False
            self._on_resume()
            self._on_resume = lambda: None

    def set_priority(self, priority: int = 0) -> None:
        """
        Control the execution priority for this effect.

        Parameters
        ----------
        priority
            The new priority. A higher value means higher priority: an effect with a
            higher priority value will execute before all effects with lower priority
            values. Positive, negative, and zero values are allowed.

        Note
        ----
        If the observer is currently invalidated, then the change in priority will not
        take effect until the next invalidation--unless the observer is also currently
        suspended, in which case the priority change will be effective upon resume.
        """
        self._priority = priority

    def _on_session_ended_cb(self) -> None:
        self.destroy()


@overload
def Effect(fn: Union[EffectFunction, EffectFunctionAsync]) -> Effect_:
    ...


@overload
def Effect(
    *,
    suspended: bool = False,
    priority: int = 0,
    session: Union[MISSING_TYPE, "Session", None] = MISSING,
) -> Callable[[Union[EffectFunction, EffectFunctionAsync]], Effect_]:
    ...


@add_example()
def Effect(
    fn: Optional[Union[EffectFunction, EffectFunctionAsync]] = None,
    *,
    suspended: bool = False,
    priority: int = 0,
    session: Union[MISSING_TYPE, "Session", None] = MISSING,
) -> Union[Effect_, Callable[[Union[EffectFunction, EffectFunctionAsync]], Effect_]]:
    """
    Mark a function as a reactive side effect.

    A reactive effect is like a reactive calculation (:func:`Calc`) in that it can read
    reactive values and call reactive calculations, and will automatically re-execute
    when those dependencies change. But unlike reactive calculations, it doesn't return
    a result and can't be used as an input to other reactive expressions. Thus,
    observers are only useful for their side effects (for example, performing I/O).

    Another contrast between reactive calculations and effects is their execution
    strategy. Reactive calculations use lazy evaluation; that is, when their
    dependencies change, they don't re-execute right away but rather wait until they are
    called by someone else. Indeed, if they are not called then they will never
    re-execute. In contrast, effects use eager evaluation; as soon as their dependencies
    change, they schedule themselves to re-execute.

    Parameters
    ----------
    suspended
        If ``TRUE``, start the effect in a suspended state (i.e., it will not execute
        until resumed and invalidated).
    priority
        The new priority. A higher value means higher priority: an effect with a higher
        priority value will execute before all effects with lower priority values.
        Positive, negative, and zero values are allowed.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    Returns
    -------
    A decorator that marks a function as a reactive effect (:class:`Effect_`).

    See Also
    --------
    ~shiny.Inputs
    ~shiny.reactive.Value
    ~shiny.reactive.Effect
    ~shiny.reactive.invalidate_later
    ~shiny.event
    """

    def create_effect(fn: Union[EffectFunction, EffectFunctionAsync]) -> Effect_:
        fn = cast(EffectFunction, fn)
        return Effect_(fn, suspended=suspended, priority=priority, session=session)

    if fn is None:
        return create_effect
    else:
        return create_effect(fn)


# ==============================================================================
# event decorator
# ==============================================================================
@add_example()
def event(
    *args: Union[Callable[[], object], Callable[[], Awaitable[object]]],
    ignore_none: bool = True,
    ignore_init: bool = False,
) -> Callable[[Callable[[], T]], Callable[[], T]]:
    """
    Mark a function to react only when an "event" occurs.

    Shiny's reactive programming framework is primarily designed for calculated values
    (:func:`~shiny.reactive.Calc`) and side-effect-causing actions
    (:func:`~shiny.reactive.Effect`) that respond to **any** of their inputs changing.
    That's often what is desired in Shiny apps, but not always: sometimes you want to
    wait for a specific action to be taken from the user, like clicking an
    :func:`~shiny.ui.input_action_button`, before calculating or taking an action. A
    reactive value (or function) which triggers other calculation or action in this way
    is called an event.

    These situations demand a more imperative, "event handling" style of programming,
    which ``@reactive.event()`` provides. It does this by using the
    :func:`~shiny.reactive.isolate` primitive under-the-hood to essentially "limit" the
    set of reactive dependencies to those in ``args``.

    Parameters
    ----------
    args
        One or more callables that represent the event; most likely this will be a
        reactive input value linked to a :func:`~shiny.ui.input_action_button` or
        similar (e.g., ``input.click``), but it can also be a (reactive or non-reactive)
        function that returns a value.
    ignore_none
        Whether to ignore the event if the value is ``None`` or ``0``.
    ignore_init
        If ``False``, the event trigger on the first run.

    Returns
    -------
    A decorator that marks a function as an event handler.

    Tip
    ----
    This decorator must be applied before the relevant reactivity decorator (i.e.,
    ``@reactive.event`` must be applied before ``@reactive.Effect``, ``@reactive.Calc``,
    ``@render.ui``, etc).
    """

    if any([not callable(arg) for arg in args]):
        raise TypeError(
            "All objects passed to event decorator must be callable.\n"
            + "If you are calling `@reactive.event(f())`, try calling `@reactive.event(f)` instead."
        )

    if len(args) == 0:
        raise TypeError(
            "`@reactive.event()` requires at least one argument, as in `@reactive.event(input.x)`.\n"
        )

    def decorator(user_fn: Callable[[], T]) -> Callable[[], T]:
        if not callable(user_fn):
            raise TypeError(
                "`@reactive.event()` must be applied to a function or Callable object.\n"
                + "It should usually be applied before `@Calc`,` @Effect`, `@output`, or `@render.xx` function.\n"
                + "In other words, `@reactive.event()` goes below the other decorators."
            )

        if isinstance(user_fn, Calc_):
            raise TypeError(
                "`@reactive.event()` must be applied before `@reactive.Calc`.\n"
                + "In other words, `@reactive.Calc` must be above `@reactive.event()`."
            )

        if isinstance(user_fn, RenderFunction):
            # At some point in the future, we may allow this condition, if we find an
            # use case. For now we'll disallow it, for simplicity.
            raise TypeError(
                "`@reactive.event()` must be applied before `@render.xx` .\n"
                + "In other words, `@render.xx` must be above `@reactive.event()`."
            )

        initialized = False

        async def trigger() -> None:
            vals: List[object] = []
            for arg in args:
                if is_async_callable(arg):
                    v = await arg()
                else:
                    v = arg()
                vals.append(v)

            nonlocal initialized
            if ignore_init and not initialized:
                initialized = True
                req(False)
            if ignore_none and all(map(_is_none_event, vals)):
                req(False)

        if is_async_callable(user_fn):

            @functools.wraps(user_fn)
            # Impossible to specify a return type here; we know T is
            # Awaitable[something] but I don't think there's a way to refer to the
            # `something`
            async def new_user_async_fn():
                await trigger()
                with isolate():
                    return await user_fn()

            return new_user_async_fn  # type: ignore

        elif any([is_async_callable(arg) for arg in args]):
            raise TypeError(
                "When decorating a synchronous function with @reactive.event(), all"
                + "arguments to @reactive.event() must be synchronous functions."
            )

        else:

            @functools.wraps(user_fn)
            def new_user_fn() -> T:
                run_coro_sync(trigger())
                with isolate():
                    return user_fn()

            return new_user_fn

    return decorator


def _is_none_event(val: object) -> bool:
    return val is None or (isinstance(val, ActionButtonValue) and val == 0)


# The code below is a test that the type checker is correctly inferring types. It should
# have some type errors as indicated. There doesn't seem to be a good way to run pyright
# and expect errors. Until that's supported, the best thing to do is uncomment the code
# below and check that the errors are highlighted in red as expected. See:
# https://github.com/microsoft/pyright/discussions/2163

# # fmt: off
# def test_calc():
#     @Calc(session=MISSING)
#     async def fas() -> int:
#         return 1

#     @Calc(session=MISSING)
#     def fs() -> int:
#         return 1

#     @Calc
#     async def fa() -> int:
#         return 1

#     @Calc
#     def f() -> int:
#         return 1

#     def test():
#         await fas()  # Should error
#         await fs()   # Should error
#         await fa()   # Should error
#         await f()    # Should error

#         fas()        # Should error
#         fs()
#         fa()         # Should error
#         f()

#     async def test_async():
#         await fas()
#         await fs()   # Should error
#         await fa()
#         await f()    # Should error

#         fas()        # Should error
#         fs()
#         fa()         # Should error
#         f()

# # fmt: off
# def test_event():
#     @event()
#     async def fas() -> int:
#         return 1

#     @event()
#     def fs() -> int:
#         return 1

#     @event()
#     async def fa() -> int:
#         return 1

#     @event()
#     def f() -> int:
#         return 1

#     def test():
#         await fas()  # Should error
#         await fs()   # Should error
#         await fa()   # Should error
#         await f()    # Should error

#         fas()        # Should error
#         fs()
#         fa()         # Should error
#         f()

#     async def test_async():
#         await fas()
#         await fs()   # Should error
#         await fa()
#         await f()    # Should error

#         fas()        # Should error
#         fs()
#         fa()         # Should error
#         f()



import random
from shiny import *

app_ui = ui.page_fluid(
    ui.markdown(
        f"""
        This example demonstrates how `@reactive.event()` can be used to restrict
        execution of: (1) a `@render` function, (2) `@reactive.Calc`, or (3)
        `@reactive.Effect`.

        In all three cases, the output is dependent on a random value that gets updated
        every 0.5 seconds (currently, it is {ui.output_ui("number", inline=True)}), but
        the output is only updated when the button is clicked.
        """
    ),
    ui.row(
        ui.column(
            3,
            ui.input_action_button("btn_out", "(1) Update number"),
            ui.output_text("out_out"),
        ),
        ui.column(
            3,
            ui.input_action_button("btn_calc", "(2) Show 1 / number"),
            ui.output_text("out_calc"),
        ),
        ui.column(
            3,
            ui.input_action_button("btn_effect", "(3) Log number"),
            ui.div(id="out_effect"),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):

    # Update a random number every second
    val = reactive.Value(random.randint(0, 1000))

    @reactive.Effect
    def _():
        reactive.invalidate_later(0.5)
        val.set(random.randint(0, 1000))

    # Always update this output when the number is updated
    @output
    @render.ui
    def number():
        return val.get()

    # Since ignore_none=False, the function executes before clicking the button.
    # (input.btn_out() is 0 on page load, but @@reactive.event() treats 0 as None for
    # action buttons.)
    @output
    @render.text
    @reactive.event(input.btn_out, ignore_none=False)
    def out_out():
        return str(val.get())

    @reactive.Calc
    @reactive.event(input.btn_calc)
    def calc():
        return 1 / val.get()

    @output
    @render.text
    def out_calc():
        return str(calc())

    @reactive.Effect
    @reactive.event(input.btn_effect)
    def _():
        ui.insert_ui(
            ui.p("Random number!", val.get()),
            selector="#out_effect",
            where="afterEnd",
        )


app = App(app_ui, server)



from shiny import *
import asyncio

app_ui = ui.page_fluid(
    ui.input_action_button("button", "Compute"),
    ui.output_text("compute"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.text
    @reactive.event(input.button)
    async def compute():
        with ui.Progress(min=1, max=15) as p:
            p.set(message="Calculation in progress", detail="This may take a while...")

            for i in range(1, 15):
                p.set(i, message="Computing")
                await asyncio.sleep(0.1)
                # Normally use time.sleep() instead, but it doesn't yet work in Pyodide.
                # https://github.com/pyodide/pyodide/issues/2354

        return "Done computing!"


app = App(app_ui, server)



"""Tests for `shiny.utils` async-related functions."""

import contextvars
import pytest
import asyncio
from typing import Iterator, List

from shiny._utils import run_coro_sync, run_coro_hybrid


def range_sync(n: int) -> Iterator[int]:
    """
    An implementation of `range()` which uses `yield`, but doesn't actually
    give up control to the event loop.
    """
    num = 0
    while num < n:
        yield num
        num += 1


async def make_list_sync(n: int) -> List[int]:
    """
    An `async` function that is in fact synchronous; it does not actually give
    up control.
    """
    x: list[int] = []
    for i in range_sync(n):
        x.append(i)
    return x


async def make_list_async(n: int) -> List[int]:
    """An `async` function that gives up control."""
    x: list[int] = []
    for i in range_sync(n):
        await asyncio.sleep(0)
        x.append(i)
    return x


def test_run_coro_sync():
    # Running a coroutine that is in fact synchronous works fine.
    res = run_coro_sync(make_list_sync(1))
    assert res == [0]

    res = run_coro_sync(make_list_sync(3))
    assert res == [0, 1, 2]

    # Should error because the asyncio.sleep() gives up control.
    with pytest.raises(RuntimeError):
        run_coro_sync(make_list_async(1))

    with pytest.raises(RuntimeError):
        run_coro_sync(make_list_async(3))

    # Same with a direct call to asyncio.sleep()
    with pytest.raises(RuntimeError):
        run_coro_sync(asyncio.sleep(0))

    with pytest.raises(RuntimeError):
        run_coro_sync(asyncio.sleep(0.1))


def test_run_coro_async():
    async def async_main():
        # awaited calls to the in-fact-synchronous function are OK.
        res = await make_list_sync(3)
        assert res == [0, 1, 2]

        # awaited calls to the async function are OK.
        res = await make_list_async(3)
        assert res == [0, 1, 2]

        await asyncio.sleep(0)

        # Calling run_coro_sync() should be the same as when called normally
        # (from a regular function, not an async function run by asyncio.run()).
        res = run_coro_sync(make_list_sync(3))
        assert res == [0, 1, 2]

        with pytest.raises(RuntimeError):
            run_coro_sync(make_list_async(3))
        with pytest.raises(RuntimeError):
            run_coro_sync(asyncio.sleep(0))

    asyncio.run(async_main())


def test_run_coro_sync_type_check():
    # Should raise an error if passed a regular generator (as opposed to a
    # coroutine object).
    with pytest.raises(TypeError):
        run_coro_sync(range_sync(0))  # type: ignore


def test_async_generator():
    # run_coro_sync() can't run async generators, but it can run async functions
    # which call async generators.

    # An async generator
    async def async_gen_range(n: int):
        for i in range(n):
            yield i

    # An async function which uses the generator
    async def main(n: int):
        x: list[int] = []
        async for i in async_gen_range(n):
            x.append(i)
        return x

    # Running the async function works fine.
    res = run_coro_sync(main(3))
    assert res == [0, 1, 2]

    # Attempting to run the async generator results in an error, because it
    # doesn't return a coroutine object.
    with pytest.raises(TypeError):
        run_coro_sync(async_gen_range(3))  # type: ignore


def test_create_task():
    # Should be OK to call create_task().
    async def create_task_wrapper():
        async def inner():
            asyncio.create_task(make_list_async(3))

        run_coro_sync(inner())

    asyncio.run(create_task_wrapper())

    # Should not be OK to await a task, because it doesn't complete immediately.
    async def create_task_wrapper2():
        async def inner():
            await asyncio.create_task(make_list_async(3))

        run_coro_sync(inner())

    with pytest.raises(RuntimeError):
        asyncio.run(create_task_wrapper2())


@pytest.mark.asyncio
async def test_coro_hybrid():
    state = 0

    async def test_task() -> int:
        nonlocal state
        state = 1
        await asyncio.sleep(0)
        state = 2
        await asyncio.sleep(0.1)
        state = 3
        return 100

    assert state == 0
    fut = run_coro_hybrid(test_task())
    assert state == 1
    await asyncio.sleep(0.01)
    assert state == 2
    await asyncio.sleep(0.1)
    assert state == 3
    assert await fut == 100


@pytest.mark.asyncio
async def test_coro_hybrid_throw():
    async def test_task_throw():
        raise ValueError("boom")

    fut = run_coro_hybrid(test_task_throw())
    with pytest.raises(ValueError):
        await fut


@pytest.mark.asyncio
async def test_coro_hybrid_throw_later():

    state = 0

    async def test_task_throw_later():
        nonlocal state
        state = 1
        await asyncio.sleep(0.1)
        raise ValueError("boom")

    fut = run_coro_hybrid(test_task_throw_later())
    assert state == 1
    with pytest.raises(ValueError):
        await fut


@pytest.mark.asyncio
async def test_coro_hybrid_cancel():
    state = 0

    async def test_task_cancel():
        nonlocal state
        state = 1
        await asyncio.sleep(0)
        state = 2

    fut = run_coro_hybrid(test_task_cancel())
    assert state == 1
    fut.cancel()
    await asyncio.sleep(0.1)
    assert state == 1


@pytest.mark.asyncio
async def test_coro_hybrid_self_cancel():
    state = 0

    fut: asyncio.Future[None]

    async def test_task_cancel():
        nonlocal state
        state = 1
        await asyncio.sleep(0)
        fut.cancel()
        await asyncio.sleep(0)
        state = 2

    fut = run_coro_hybrid(test_task_cancel())
    assert state == 1
    await asyncio.sleep(0.1)
    assert state == 1
    with pytest.raises(asyncio.CancelledError):
        await fut
    assert state == 1


@pytest.mark.asyncio
async def test_coro_hybrid_self_cancel2():
    state = 0

    async def test_task_cancel():
        nonlocal state
        state = 1
        await asyncio.sleep(0)
        raise asyncio.CancelledError()

    fut = run_coro_hybrid(test_task_cancel())
    assert state == 1
    await asyncio.sleep(0.1)
    assert state == 1
    with pytest.raises(asyncio.CancelledError):
        await fut
    assert state == 1


@pytest.mark.asyncio
async def test_coro_hybrid_context():
    test = contextvars.ContextVar("test", default=False)

    async def test_ctx():
        assert test.get()
        await asyncio.sleep(0)
        assert test.get()
        await asyncio.sleep(0.1)
        assert test.get()

    def do():
        # Set our context's copy of `test` to True. Since test_ctx() is being launched
        # within our context, it should see test.get() == True, whereas any code outside
        # our context will continue to see test.get() == False.
        test.set(True)
        return run_coro_hybrid(test_ctx())

    inner_ctx = contextvars.copy_context()
    fut = inner_ctx.run(do)
    assert not test.get()
    await fut
    assert not test.get()



from ._docstring import add_example
from .types import SilentException, SilentCancelOutputException


@add_example()
def req(*args: object, cancel_output: bool = False) -> None:
    """
    Throw a silent exception for falsey values.

    This is a convenient shorthand for throwing :func:`~shiny.types.SilentException` /
    :func:`~shiny.types.SilentCancelOutputException` if any of the arguments are falsey.

    Parameters
    ----------
    args
        Any number of arguments to check.
    cancel_output
        If ``True``, throw :func:`~shiny.types.SilentCancelOutputException` instead of
        :func:`~shiny.types.SilentException`.
    """
    for arg in args:
        if not arg:
            if cancel_output:
                raise SilentCancelOutputException()
            else:
                raise SilentException()



import os
import sys
from typing import Callable, Any, TypeVar, List

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


ex_dir: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), "examples")

FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)

# This class is used to mark docstrings when @add_example() is used, so that an error
# will be thrown if @doc_format() is used afterward. This is to avoid an error when
# the example contains curly braces -- the @doc_format() decorator will try to evaluate
# the code in {}.
class DocStringWithExample(str):
    ...


def add_example(
    directive: Literal[
        "shinyapp::", "shinyeditor::", "code-block:: python", "cell::", "terminal::"
    ] = "shinyeditor::",
    **options: str,
) -> Callable[[F], F]:
    """
    Add an example to the docstring of a function, method, or class.

    This decorator must, at the moment, be used on a function, method, or class whose
    ``__name__`` matches the name of directory under ``shiny/examples/``, and must
    also contain a ``app.py`` file in that directory.

    Parameters
    ----------
    directive
        A directive for rendering the example. This can be one of:
            - ``shinyapp``: A live shiny app (statically served via wasm).
            - ``code``: A python code snippet.
            - ``shinyeditor``: A live shiny app (statically served via wasm).
            - ``cell``: A executable Python cell.
            - ``terminal``: A minimal Python IDE
    **options
        Options for the directive. See docs/source/sphinxext/pyshinyapp.py for details.
    """

    def _(func: F) -> F:

        # To avoid a performance hit on `import shiny`, we only add examples to the
        # docstrings if this env variable is set (as it is in docs/source/conf.py).
        if os.getenv("SHINY_ADD_EXAMPLES") != "true":
            if func.__doc__ is not None:
                func.__doc__ = DocStringWithExample(func.__doc__)
            return func

        fn_name = func.__name__
        example_file = os.path.join(ex_dir, fn_name, "app.py")
        if not os.path.exists(example_file):
            raise ValueError(f"No example for {fn_name}")

        if func.__doc__ is None:
            func.__doc__ = ""

        # How many leading spaces does the docstring start with?
        doc = func.__doc__.replace("\n", "")
        indent = " " * (len(doc) - len(doc.lstrip()))

        with open(example_file) as f:
            example = indent.join([" " * 4 + x for x in f.readlines()])

        # When rendering a standalone app, put the code above it (maybe this should be
        # handled by the directive itself?)
        example_prefix: List[str] = []
        if directive == "shinyapp::":
            example_prefix.extend(
                [
                    ".. code-block:: python",
                    "",
                    example,
                    "",
                ]
            )

        example_section = ("\n" + indent).join(
            [
                "",
                "",
                "Example",
                "-------",
                "",
                *example_prefix,
                f".. {directive}",
                *[f"{indent}:{k}: {v}" for k, v in options.items()],
                "",
                example,
            ]
        )

        func.__doc__ += example_section
        func.__doc__ = DocStringWithExample(func.__doc__)
        return func

    return _


def doc_format(**kwargs: str) -> Callable[[F], F]:
    def _(func: F) -> F:
        if isinstance(func.__doc__, DocStringWithExample):
            raise ValueError(
                f"@doc_format() must be applied before @add_example() for {func.__name__}."
            )
        if func.__doc__:
            func.__doc__ = func.__doc__.format(**kwargs)
        return func

    return _



"""
Tools for working within a (user) session context.
"""
from ._session import Session, Inputs, Outputs
from ._utils import (  # noqa: F401
    get_current_session,
    session_context,  # pyright: ignore[reportUnusedImport]
    require_active_session,
)

__all__ = (
    "Session",
    "Inputs",
    "Outputs",
    "get_current_session",
    "require_active_session",
)



__all__ = ("notification_show", "notification_remove")

import sys
from typing import Dict, Union, Optional, Any

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from htmltools import TagList, TagChildArg

from .._docstring import add_example
from .._utils import rand_hex
from ..session import Session, require_active_session


@add_example()
def notification_show(
    ui: TagChildArg,
    action: Optional[TagList] = None,
    duration: Optional[Union[int, float]] = 5,
    close_button: bool = True,
    id: Optional[str] = None,
    type: Literal["default", "message", "warning", "error"] = "default",
    session: Optional[Session] = None,
) -> str:
    """
    Show a notification to the user.

    Parameters
    ----------
    ui
        Content of message.
    action
        Message content that represents an action. For example, this could be a link
        that the user can click on. This is separate from ui so customized layouts can
        handle the main notification content separately from action content.
    duration
        Number of seconds to display the message before it disappears. Use ``None`` to
        make the message not automatically disappear.
    close_button
        If ``True``, display a button which will make the notification disappear when
        clicked. If ``False`` do not display.
    id
        An optional unique identifier for the notification. If supplied, any existing
        notification with the same ``id`` will be replaced with this one (otherwise, a
        new notification is created).
    type
        A string which controls the color of the notification. One of "default" (gray),
        "message" (blue), "warning" (yellow), or "error" (red).
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    Returns
    -------
    The notification's ``id``.

    See Also
    -------
    ~shiny.ui.notification_remove
    ~shiny.ui.modal
    """

    session = require_active_session(session)

    ui_ = session._process_ui(ui)
    action_ = session._process_ui(action)

    id = id if id else rand_hex(8)

    payload: Dict[str, Any] = {
        "html": ui_["html"],
        "action": action_["html"],
        "deps": ui_["deps"] + action_["deps"],
        "closeButton": close_button,
        "id": id,
        "type": type,
    }

    if duration:
        payload.update({"duration": duration * 1000})

    session._send_message_sync({"notification": {"type": "show", "message": payload}})

    return id


def notification_remove(id: str, session: Optional[Session] = None) -> str:
    """
    Remove a notification.

    Parameters
    ----------
    id
        A notification ``id``.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
       :func:`~shiny.session.get_current_session`.

    Returns
    -------
    The notification's ``id``.

    See Also
    -------
    ~shiny.ui.notification_show
    ~shiny.ui.modal

    Example
    -------
    See :func:`notification_show`.
    """
    session = require_active_session(session)
    session._send_message_sync({"notification": {"type": "remove", "message": id}})
    return id



from shiny import *

app_ui = ui.page_fluid(
    ui.input_selectize(
        "state",
        "Choose a state:",
        {
            "East Coast": {"NY": "NY", "NJ": "NJ", "CT": "CT"},
            "West Coast": {"WA": "WA", "OR": "OR", "CA": "CA"},
            "Midwest": {"MN": "MN", "WI": "WI", "IA": "IA"},
        },
        multiple=True,
    ),
    ui.output_text("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.text
    def value():
        return "You choose: " + str(input.state())


app = App(app_ui, server)



import base64
import io
import os
import sys
from typing import Optional, Union, cast, Tuple, TextIO, BinaryIO, Any, List

if sys.version_info >= (3, 8):
    from typing import Literal, Protocol
else:
    from typing_extensions import Literal, Protocol

from ..types import ImgData

# Use this protocol to avoid needing to maintain working stubs for matplotlib. If
# good stubs ever become available for matplotlib, use those instead.
class MplFigure(Protocol):
    def set_size_inches(
        self,
        w: Union[Tuple[float, float], float],
        h: Optional[float] = None,
        forward: bool = True,
    ):
        ...

    def savefig(
        self,
        fname: Union[str, TextIO, BinaryIO, "os.PathLike[Any]"],
        dpi: Union[float, Literal["figure"], None] = None,
        # facecolor="w",
        # edgecolor="w",
        # orientation="portrait",
        # papertype=None,
        format: Optional[str] = None,
        # transparent=False,
        bbox_inches: object = None,
        # pad_inches=0.1,
        # frameon=None,
        # metadata=None,
    ):
        ...


class MplArtist(Protocol):
    def get_figure(self) -> MplFigure:
        ...


class MplAnimation(Protocol):
    def pause(self):
        ...

    def resume(self):
        ...


# Use this protocol to avoid needing to maintain working stubs for plotnint. If
# good stubs ever become available for plotnine, use those instead.
class PlotnineFigure(Protocol):
    def save(
        self,
        filename: BinaryIO,
        format: str,
        units: str,
        dpi: float,
        width: float,
        height: float,
        verbose: bool,
        bbox_inches: object = None,
    ):
        ...


TryPlotResult = Union[ImgData, None, Literal["TYPE_MISMATCH"]]


# Try to render a matplotlib object. If `fig` is not a matplotlib object, return
# "TYPE_MISMATCH". If there's an error in rendering, return None. If successful in
# rendering, return an ImgData object.
def try_render_matplotlib(
    x: object,
    width: float,
    height: float,
    pixelratio: float,
    ppi: float,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    fig = get_matplotlib_figure(x)

    if fig is None:
        return "TYPE_MISMATCH"

    try:
        fig.set_size_inches(width / ppi, height / ppi)

        bbox_inches = kwargs.pop("bbox_inches", "tight")

        with io.BytesIO() as buf:
            fig.savefig(
                buf,
                format="png",
                dpi=ppi * pixelratio,
                bbox_inches=bbox_inches,
                **kwargs,
            )
            buf.seek(0)
            data = base64.b64encode(buf.read())
            data_str = data.decode("utf-8")

        # N.B. matplotlib.tight_layout() causes the intrinsic file size can be different
        # from the requested size (i.e., the container size). So, scale the image to fit
        # in the container while preserving the aspect ratio.
        res: ImgData = {
            "src": "data:image/png;base64," + data_str,
            "width": "100%",
            "height": "100%",
            "style": "object-fit:contain",
        }

        if alt is not None:
            res["alt"] = alt

        return res

    except Exception as e:
        # TODO: just let errors propagate?
        print("Error rendering matplotlib object: " + str(e))

    finally:
        import matplotlib.pyplot  # pyright: ignore[reportMissingTypeStubs]

        matplotlib.pyplot.close(fig)  # pyright: ignore[reportUnknownMemberType]

    return None


def get_matplotlib_figure(x: object) -> Union[MplFigure, None]:
    from matplotlib.figure import (  # pyright: reportMissingTypeStubs=false,reportUnknownVariableType=false
        Figure,
    )
    from matplotlib.artist import (  # pyright: reportMissingTypeStubs=false,reportUnknownVariableType=false
        Artist,
    )
    from matplotlib.animation import (  # pyright: reportMissingTypeStubs=false,reportUnknownVariableType=false
        Animation,
    )

    if isinstance(x, Figure):
        return cast(MplFigure, x)

    if isinstance(x, Animation):
        raise RuntimeError(
            "Matplotlib's Animation class isn't supported by @render.plot. "
            + "Consider explictly saving the animation to a file and "
            + "then using @render.image instead to render it."
        )

    # Libraries like pandas, xarray, etc have plot() methods that can return a wide
    # array pf mpl classes, like Lines2D, Subplots, Axes, etc. The Artist ABC class
    # should cover most, if not all, of these (it doesn't cover Animation, though).
    # https://matplotlib.org/stable/api/artist_api.html
    if isinstance(x, Artist):
        return cast(MplArtist, x).get_figure()

    # Some other custom figure-like classes such as seaborn.axisgrid.FacetGrid attach
    # their figure as an attribute
    fig = getattr(x, "figure", None)
    if isinstance(fig, Figure):
        return cast(MplFigure, fig)

    # Sometimes generic plot() methods will return an iterable of Artists,
    # If they all refer to the same figure, then it seems reasonable to use it
    # https://docs.xarray.dev/en/latest/user-guide/plotting.html#dimension-along-y-axis
    if isinstance(x, (list, tuple)):
        figs = [get_matplotlib_figure(y) for y in cast(List[Any], x)]
        if len(set(figs)) == 1:
            return figs[0]

    return None


def try_render_pil(
    x: object,
    width: float,
    height: float,
    pixelratio: float,
    ppi: float,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    import PIL.Image  # pyright: ignore[reportMissingModuleSource]

    if not isinstance(x, PIL.Image.Image):
        return "TYPE_MISMATCH"

    try:
        with io.BytesIO() as buf:
            x.save(buf, format="PNG", **kwargs)
            buf.seek(0)
            data = base64.b64encode(buf.read())
            data_str = data.decode("utf-8")

        res: ImgData = {
            "src": "data:image/png;base64," + data_str,
            "width": "100%",
            "height": "100%",
            "style": "object-fit:contain",
        }

        if alt is not None:
            res["alt"] = alt

        return res

    except Exception as e:
        # TODO: just let errors propagate?
        print("Error rendering PIL object: " + str(e))

    return None


def try_render_plotnine(
    x: object,
    width: float,
    height: float,
    pixelratio: float,
    ppi: float,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    from plotnine.ggplot import (  # pyright: reportMissingTypeStubs=false,reportUnknownVariableType=false,reportMissingImports=false
        ggplot,
    )

    if not isinstance(x, ggplot):
        return "TYPE_MISMATCH"

    bbox_inches = kwargs.pop("bbox_inches", "tight")
    try:
        with io.BytesIO() as buf:
            cast(PlotnineFigure, x).save(
                filename=buf,
                format="png",
                units="in",
                dpi=ppi * pixelratio,
                width=width / ppi,
                height=height / ppi,
                verbose=False,
                bbox_inches=bbox_inches,
                **kwargs,
            )
            buf.seek(0)
            data = base64.b64encode(buf.read())
            data_str = data.decode("utf-8")

        res: ImgData = {
            "src": "data:image/png;base64," + data_str,
            "width": "100%",
            "height": "100%",
            "style": "object-fit:contain",
        }

        if alt is not None:
            res["alt"] = alt

        return res

    except Exception as e:
        # TODO: just let errors propagate?
        print("Error rendering matplotlib object: " + str(e))

    return None



from shiny import *

app_ui = ui.page_fluid(
    ui.input_checkbox_group(
        "colors",
        "Choose color(s):",
        {
            "red": ui.span("Red", style="color: #FF0000;"),
            "green": ui.span("Green", style="color: #00AA00;"),
            "blue": ui.span("Blue", style="color: #0000AA;"),
        },
    ),
    ui.output_ui("val"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    def val():
        req(input.colors())
        return "You chose " + ", ".join(input.colors())


app = App(app_ui, server)



from shiny import *
from shiny.types import SafeException

app_ui = ui.page_fluid(ui.output_ui("safe"), ui.output_ui("unsafe"))


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    def safe():
        raise SafeException("This is a safe exception")

    @output
    @render.ui
    def unsafe():
        raise Exception("This is an unsafe exception")


app = App(app_ui, server)
app.sanitize_errors = True



from shiny import *
from datetime import date

app_ui = ui.page_fluid(
    ui.input_date_range(
        "daterange1", "Date range:", start="2001-01-01", end="2010-12-31"
    ),
    # Default start and end is the current date in the client's time zone
    ui.input_date_range("daterange2", "Date range:"),
    # start and end are always specified in yyyy-mm-dd, even if the display
    # format is different
    ui.input_date_range(
        "daterange3",
        "Date range:",
        start="2001-01-01",
        end="2010-12-31",
        min="2001-01-01",
        max="2012-12-21",
        format="mm/dd/yy",
        separator=" - ",
    ),
    # Pass in Date objects
    ui.input_date_range(
        "daterange4", "Date range:", start=date(2001, 1, 1), end=date(2010, 12, 31)
    ),
    # Use different language and different first day of week
    ui.input_date_range("daterange5", "Date range:", language="de", weekstart=1),
    # Start with decade view instead of default month view
    ui.input_date_range("daterange6", "Date range:", startview="decade"),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



from typing import List

from shiny import *
from shiny.types import NavSetArg
from shiny.ui import h4


def nav_controls(prefix: str) -> List[NavSetArg]:
    return [
        ui.nav("a", prefix + ": tab a content"),
        ui.nav("b", prefix + ": tab b content"),
        ui.nav_control(
            ui.a(
                "Shiny",
                href="https://github.com/rstudio/shiny",
                target="_blank",
            )
        ),
        ui.nav_spacer(),
        ui.nav_menu(
            "Other links",
            ui.nav("c", prefix + ": tab c content"),
            "----",
            "Plain text",
            "----",
            ui.nav_control(
                ui.a(
                    "RStudio",
                    href="https://rstudio.com",
                    target="_blank",
                )
            ),
            align="right",
        ),
    ]


app_ui = ui.page_navbar(
    *nav_controls("page_navbar"),
    title="page_navbar()",
    bg="#0062cc",
    inverse=True,
    id="navbar_id",
    footer=ui.div(
        {"style": "width:80%;margin: 0 auto"},
        ui.tags.style(
            """
            h4 {
                margin-top: 3em;
            }
            """
        ),
        h4("navset_tab()"),
        ui.navset_tab(*nav_controls("navset_tab()")),
        h4("navset_pill()"),
        ui.navset_pill(*nav_controls("navset_pill()")),
        h4("navset_tab_card()"),
        ui.navset_tab_card(*nav_controls("navset_tab_card()")),
        h4("navset_pill_card()"),
        ui.navset_pill_card(*nav_controls("navset_pill_card()")),
        h4("navset_pill_list()"),
        ui.navset_pill_list(*nav_controls("navset_pill_list()")),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        print("Current navbar page: ", input.navbar_id())


app = App(app_ui, server)



from shiny import *
from shiny.types import FileInfo
import pandas as pd

app_ui = ui.page_fluid(
    ui.layout_sidebar(
        ui.panel_sidebar(
            ui.input_file("file1", "Choose CSV File", accept=[".csv"], multiple=False),
            ui.input_checkbox("header", "Header", True),
        ),
        ui.panel_main(ui.output_ui("contents")),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    def contents():
        if input.file1() is None:
            return "Please upload a csv file"
        f: list[FileInfo] = input.file1()
        df = pd.read_csv(f[0]["datapath"], header=0 if input.header() else None)
        return ui.HTML(df.to_html(classes="table table-striped"))


app = App(app_ui, server)
